"""
babel_script.py last modified 2014-04-04 13:55:07
Original script by Yourself
Anti grief by izzy
Return intel dropped from platform bug fix by a_girl

Release thread:
http://www.buildandshoot.com/viewtopic.php?t=2586

How to install and configure:

1) Save babel_script.py to 'scripts' folder:
http://aloha.pk/files/aos/pyspades/feature_server/scripts/babel_script.py
2) Save babel.py to 'scripts' folder:
http://aloha.pk/files/aos/pyspades/feature_server/scripts/babel.py
3) Set game_mode to "babel" in config.txt
4) Add "babel_script" to scripts list in config.txt
5) Set cap_limit to "10" in config.txt
"""

from random import randint
from pyspades.constants import (BLUE_BASE, GREEN_BASE, BLUE_FLAG, GREEN_FLAG,
                                SPADE_TOOL, GRENADE_TOOL, WEAPON_TOOL)
from twisted.internet import reactor

# If ALWAYS_ENABLED is False, then babel can be enabled by setting 'babel': True
# in the map metadat extensions dictionary.
ALWAYS_ENABLED = True

PLATFORM_WIDTH = 100
PLATFORM_HEIGHT = 32
PLATFORM_COLOR = (0, 255, 255, 255)
BLUE_BASE_COORDS = (256 - 138, 256)
GREEN_BASE_COORDS = (256 + 138, 256)
SPAWN_SIZE = 40

# Don't touch this stuff
PLATFORM_WIDTH //= 2
PLATFORM_HEIGHT //= 2
SPAWN_SIZE //= 2


def get_entity_location(self, entity_id):
    if entity_id == BLUE_BASE:
        return BLUE_BASE_COORDS + (self.protocol.map.get_z(*BLUE_BASE_COORDS),)
    elif entity_id == GREEN_BASE:
        return GREEN_BASE_COORDS + \
            (self.protocol.map.get_z(*GREEN_BASE_COORDS),)
    elif entity_id == BLUE_FLAG:
        return (256 - PLATFORM_WIDTH + 1, 256, 0)
    elif entity_id == GREEN_FLAG:
        return (256 + PLATFORM_WIDTH - 1, 256, 0)


def get_spawn_location(connection):
    xb = connection.team.base.x
    yb = connection.team.base.y
    xb += randint(-SPAWN_SIZE, SPAWN_SIZE)
    yb += randint(-SPAWN_SIZE, SPAWN_SIZE)
    return (xb, yb, connection.protocol.map.get_z(xb, yb))


def coord_on_platform(x, y, z):
    if z <= 2:
        if (256 - PLATFORM_WIDTH <= x <= 256 + PLATFORM_WIDTH and
                256 - PLATFORM_HEIGHT <= y <= 256 + PLATFORM_HEIGHT):
            return True
    if z == 1:
        if (256 - PLATFORM_WIDTH - 1 <= x <= 256 + PLATFORM_WIDTH + 1
            and 256 - PLATFORM_HEIGHT - 1 <= y <= 256 + PLATFORM_HEIGHT + 1):
            return True
    return False

def apply_script(protocol, connection, config):
    allowed_intel_hold_time = config.get('allowed_intel_hold_time', 150)

    class TowerOfBabelConnection(connection):
        def get_protected_area(self, team):
            """returns minx, maxx, miny, maxy"""
            if team is self.protocol.blue_team:
                return 301, 384, 240, 272
            else:
                return 128, 211, 240, 272

        def invalid_build_position(self, x, y, z):
            position = self.world_object.position

            if not self.god and self.protocol.babel:
                if coord_on_platform(x, y, z):
                    connection.on_block_build_attempt(self, x, y, z)
                    return True
            # prevent enemies from building in protected areas

            minx, maxx, miny, maxy = self.get_protected_area(self.team)
            if minx <= position.x <= maxx and miny <= position.y <= maxy:
                self.send_chat('You can\'t build near the enemy\'s tower!')
                return True
            return False

        def on_block_build_attempt(self, x, y, z):
            if self.invalid_build_position(x, y, z):
                return False
            return connection.on_block_build_attempt(self, x, y, z)

        def on_line_build_attempt(self, points):
            for point in points:
                if self.invalid_build_position(*point):
                    return False
            return connection.on_line_build_attempt(self, points)

        def is_trusted_for_block_destruction(self):
            return self.admin or self.user_types.moderator or self.user_types.guard or self.user_types.trusted

        # anti team destruction
        def on_block_destroy(self, x, y, z, mode):
            minx, maxx, miny, maxy = self.get_protected_area(self.team.other)

            position = self.world_object.position

            if not self.is_trusted_for_block_destruction() and self.tool is SPADE_TOOL:
                if minx <= position.x <= maxx and miny <= position.y <= maxy:
                    self.send_chat('You can\'t destroy your team\'s blocks in this area. Attack the enemy\'s tower!')
                    return False

            if self.team is self.protocol.blue_team:
                can_shoot_blocks = position.x <= 288
            else:
                can_shoot_blocks = position.x >= 224

            if can_shoot_blocks:
                if self.tool is WEAPON_TOOL:
                    self.send_chat('You must be closer to the enemy\'s base to shoot blocks!')
                else:
                    self.send_chat('You must be closer to the enemy\'s base to grenade blocks!')
                return False
            return connection.on_block_destroy(self, x, y, z, mode)

        auto_kill_intel_hog_call = None

        # kill intel carrier if held too long
        def auto_kill_intel_hog(self):
            self.auto_kill_intel_hog_call = None
            self.kill()
            self.protocol.send_chat(
                'God punished %s for holding the intel too long' %
                (self.name))

        def restore_default_fog_color(self):
            self.protocol.set_fog_color(
                getattr(self.protocol.map_info.info, 'fog', (128, 232, 255)))

        def on_flag_take(self):
            if self.auto_kill_intel_hog_call is not None:
                self.auto_kill_intel_hog_call.cancel()
                self.auto_kill_intel_hog_call = None
            self.auto_kill_intel_hog_call = reactor.callLater(
                allowed_intel_hold_time, self.auto_kill_intel_hog)
            # flash team color in sky
            if self.team is self.protocol.blue_team:
                self.protocol.set_fog_color(
                    getattr(self.protocol.map_info.info, 'fog', (0, 0, 255)))
            if self.team is self.protocol.green_team:
                self.protocol.set_fog_color(
                    getattr(self.protocol.map_info.info, 'fog', (0, 255, 0)))
            reactor.callLater(0.25, self.restore_default_fog_color)
            return connection.on_flag_take(self)

        # return intel to platform if dropped
        def on_flag_drop(self):
            x, y, z = self.world_object.position.x, self.world_object.position.y, self.world_object.position.z
            if self.auto_kill_intel_hog_call is not None:
                self.auto_kill_intel_hog_call.cancel()
                self.auto_kill_intel_hog_call = None
            if z >= 0:
                self.reset_flag()
            elif (x >= (256 + PLATFORM_WIDTH)) or (x < (256 - PLATFORM_WIDTH)):
                self.reset_flag()
            elif (y >= (256 + PLATFORM_HEIGHT)) or (y < (256 - PLATFORM_HEIGHT)):
                self.reset_flag()
            self.protocol.set_fog_color(
                getattr(self.protocol.map_info.info, 'fog', (255, 0, 0)))
            reactor.callLater(0.25, self.restore_default_fog_color)
            return connection.on_flag_drop(self)

        def reset_flag(self):
            self.protocol.onectf_reset_flags()
            self.protocol.send_chat('The intel has returned to the heavens')

        def on_flag_capture(self):
            if self.auto_kill_intel_hog_call is not None:
                self.auto_kill_intel_hog_call.cancel()
                self.auto_kill_intel_hog_call = None
            return connection.on_flag_capture(self)

        def on_reset(self):
            if self.auto_kill_intel_hog_call is not None:
                self.auto_kill_intel_hog_call.cancel()
                self.auto_kill_intel_hog_call = None
            return connection.on_reset(self)

    class TowerOfBabelProtocol(protocol):
        babel = False

        def on_map_change(self, map):
            extensions = self.map_info.extensions
            if ALWAYS_ENABLED:
                self.babel = True
            else:
                if "babel" in extensions:
                    self.babel = extensions['babel']
                else:
                    self.babel = False
            if self.babel:
                self.map_info.cap_limit = 1
                self.map_info.get_entity_location = get_entity_location
                self.map_info.get_spawn_location = get_spawn_location
                for x in range(256 - PLATFORM_WIDTH, 256 + PLATFORM_WIDTH):
                    for y in range(
                            256 - PLATFORM_HEIGHT,
                            256 + PLATFORM_HEIGHT):
                        map.set_point(x, y, 1, PLATFORM_COLOR)
            return protocol.on_map_change(self, map)

        def is_indestructable(self, x, y, z):
            if self.babel:
                if coord_on_platform(x, y, z):
                    protocol.is_indestructable(self, x, y, z)
                    return True
            return protocol.is_indestructable(self, x, y, z)

    return TowerOfBabelProtocol, TowerOfBabelConnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import json

from twisted.internet.protocol import (ReconnectingClientFactory,
                                       ServerFactory)
from twisted.internet import reactor
from twisted.internet.defer import Deferred, succeed
from twisted.protocols.basic import Int16StringReceiver

CONNECTION_TIMEOUT = 5
DEFAULT_PORT = 32880


def hash_password(value):
    return value  # hashlib.md5(value).hexdigest()


class StatsProtocol(Int16StringReceiver):

    def stringReceived(self, data):
        self.object_received(json.loads(data.decode()))

    def send_object(self, obj):
        if self.transport is not None:
            self.sendString(json.dumps(obj).encode())

    def object_received(self, obj):
        pass


class StatsServer(StatsProtocol):

    # TODO: add __init__ method for defined attributes and don't forget to call
    # super
    # pylint: disable=attribute-defined-outside-init

    def connectionMade(self):
        self.timeout_call = reactor.callLater(CONNECTION_TIMEOUT,
                                              self.timed_out)

    # TODO: check why the below
    # pylint: disable=signature-differs
    def connectionLost(self, reason):
        if self in self.factory.connections:
            self.factory.connections.remove(self)

    def object_received(self, obj):
        obj_type = obj.get('type', None)
        if self.timeout_call is not None:
            if obj_type != 'auth':
                self.transport.loseConnection()
                return
            password = obj['password']
            if password != self.factory.password:
                self.transport.loseConnection()
            else:
                self.timeout_call.cancel()
                self.timeout_call = None
                self.send_object({'type': 'authed'})
                self.name = obj['name']
                self.factory.connections.append(self)
                self.connection_accepted()
            return
        if obj_type == 'kill':
            self.add_kill(obj['name'])
        elif obj_type == 'death':
            self.add_death(obj['name'])
        elif obj_type == 'login':
            # TODO: do we want to be returning the result?
            # pylint: disable=unused-variable
            result = self.check_user(obj['name'], obj['password']).addCallback(
                self.send_login_result)

    def send_login_result(self, result):
        self.send_object({'type': 'login', 'result': result})

    def add_kill(self, name):
        pass

    def add_death(self, name):
        pass

    def check_user(self, name, password):
        pass

    def connection_accepted(self):
        pass

    def timed_out(self):
        self.timeout_call = None
        self.transport.loseConnection()


class StatsFactory(ServerFactory):
    protocol = StatsServer

    def __init__(self, password):
        self.password = hash_password(password)
        self.connections = []


class StatsClient(StatsProtocol):
    server = None
    login_defers = None

    def connectionMade(self):
        self.login_defers = []
        self.send_object({'type': 'auth', 'name': self.factory.name,
                          'password': self.factory.password})

    def object_received(self, obj):
        obj_type = obj['type']
        if obj_type == 'authed':
            self.factory.callback(self)
        elif obj_type == 'login':
            defer = self.login_defers.pop(0)
            defer.callback(obj['result'])

    def add_kill(self, name):
        self.send_object({'type': 'kill', 'name': name})

    def add_death(self, name):
        self.send_object({'type': 'death', 'name': name})

    def login_user(self, name, password):
        defer = Deferred()
        password = hash_password(password)
        self.send_object({'type': 'login', 'name': name,
                          'password': password})
        self.login_defers.append(defer)
        return defer


class StatsClientFactory(ReconnectingClientFactory):
    protocol = StatsClient
    maxDelay = 20

    def __init__(self, name, password, callback):
        self.name = name
        self.password = hash_password(password)
        self.callback = callback


def connect_statistics(host, port, name, password, callback, interface=''):
    reactor.connectTCP(host, port, StatsClientFactory(name, password, callback),
                       bindAddress=(interface, 0))

if __name__ == '__main__':
    class TestServer(StatsServer):

        def add_kill(self, name):
            print('Adding kill to', name)

        def add_death(self, name):
            print('Adding death to', name)

        def check_user(self, name, password):
            print('Checking user name/pass ({}, {})'.format(name, password))
            # TODO: pylint thinks this function should have an argument
            #       check it out
            return succeed()  # pylint: disable=no-value-for-parameter

    class TestFactory(StatsFactory):
        protocol = TestServer

    reactor.listenTCP(DEFAULT_PORT, TestFactory('marmelade'))
    reactor.run()



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import re
import random
from itertools import groupby, chain
from operator import attrgetter
from typing import List

from twisted.words.protocols import irc
from twisted.internet import reactor, protocol
from twisted.logger import Logger

from pyspades.constants import MAX_CHAT_SIZE
from pyspades.common import encode, escape_control_codes
from pyspades.types import AttributeSet
from piqueserver import commands
from piqueserver.commands import command, restrict


MAX_IRC_CHAT_SIZE = MAX_CHAT_SIZE * 2
IRC_TEAM_COLORS = {0: '\x0302', 1: '\x0303'}
SPLIT_WHO_IN_TEAMS = True
SPLIT_THRESHOLD = 20  # players

irc_color_codes = re.compile(r"\x03(?:\d{1,2}(?:,\d{1,2})?)?", re.UNICODE)

log = Logger()

def channel(func):
    """This decorator rewrites the username of incoming messages to strip the
    ident and rejects it if the source channel is not equal to the channel the
    bot is in"""
    def new_func(self, user, irc_channel, *arg, **kw):
        if not irc_channel.lower() == self.factory.channel:
            return
        user = user.split('!', 1)[0]
        func(self, user, irc_channel, *arg, **kw)
    return new_func


class IRCBot(irc.IRCClient):
    ops = None
    voices = None
    unaliased_name = None
    name = None

    @property
    def nickname(self):
        return self.factory.nickname

    @nickname.setter
    def nickname(self, nickname):
        self.factory.nickname = nickname

    @property
    def colors(self):
        return self.factory.colors

    @colors.setter
    def colors(self, colors):
        self.factory.colors = colors

    @property
    def admin(self):
        return self.factory.admin

    @property
    def user_types(self):
        return self.factory.user_types

    @property
    def rights(self):
        return self.factory.rights

    def signedOn(self):
        self.join(self.factory.channel, self.factory.password)

    def joined(self, irc_channel):
        if irc_channel.lower() == self.factory.channel:
            self.ops = set()
            self.voices = set()
        log.info("Joined channel %s" % irc_channel)

    def irc_NICK(self, prefix, params):
        user = prefix.split('!', 1)[0]
        new_user = params[0]
        if user in self.ops:
            self.ops.discard(user)
            self.ops.add(new_user)
        if user in self.voices:
            self.voices.discard(user)
            self.voices.add(new_user)

    def irc_RPL_NAMREPLY(self, *arg):
        if not arg[1][2].lower() == self.factory.channel:
            return
        for name in arg[1][3].split():
            mode = name[0]
            # reset ops or voices in case not previously set (they should be in
            # the joined method, but joined() not observed to always get called)
            if not self.ops: self.ops = set()
            if not self.voices: self.voices = set()
            l = {'@': self.ops, '+': self.voices}
            if mode in l:
                l[mode].add(name[1:])

    def left(self, irc_channel):
        if irc_channel.lower() == self.factory.channel:
            self.ops = None
            self.voices = None

    @channel
    def modeChanged(self, user, irc_channel, set_something, modes, args):
        ll = {'o': self.ops, 'v': self.voices}
        for i in range(len(args)):
            mode, name = modes[i], args[i]
            if mode not in ll:
                continue
            l = ll[mode]
            if set_something:
                l.add(name)
            elif not set_something:
                l.discard(name)

    @channel
    def privmsg(self, user, irc_channel, msg):
        if user not in self.ops and user not in self.voices:
            return  # This user is unpriviledged

        prefix = '@' if user in self.ops else '+'
        alias = self.factory.aliases.get(user, user)

        if msg.startswith(self.factory.chatprefix):
            max_len = MAX_IRC_CHAT_SIZE - \
                len(self.protocol.server_prefix) - 1
            msg = msg[len(self.factory.chatprefix):].strip()
            message = ("[irc] <{}> {}".format(prefix + alias, msg))[:max_len]
            log.info(escape_control_codes(message))
            self.factory.server.broadcast_chat(message)
        elif msg.startswith(self.factory.commandprefix) and user in self.ops:
            self.unaliased_name = user
            self.name = prefix + alias
            user_input = msg[len(self.factory.commandprefix):]
            result = commands.handle_input(self, user_input)
            if result is not None:
                self.send("{}: {}".format(user, result))

    @channel
    def userLeft(self, user, irc_channel):
        self.ops.discard(user)
        self.voices.discard(user)

    def userQuit(self, user, message):
        self.userLeft(user, self.factory.channel)

    def userKicked(self, kickee, irc_channel, kicker, message):
        self.userLeft(kickee, irc_channel)

    def send(self, msg, do_filter=False):
        if do_filter:
            msg = irc_color_codes.sub('', msg)
        self.msg(self.factory.channel, msg)

    def me(self, msg, do_filter=False):
        if do_filter:
            msg = irc_color_codes.sub('', msg)
        self.describe(self.factory.channel, msg)

    # methods used to emulate the behaviour of regular Connection objects to
    # prevent errors when command writers didn't test that their commands would
    # work when run from IRC
    def send_chat(self, value: str, _):
        self.send(value)

    def send_lines(self, lines: List[str]):
        self.send("\n".join(lines))

class IRCClientFactory(protocol.ClientFactory):
    protocol = IRCBot
    lost_reconnect_delay = 20
    failed_reconnect_delay = 60
    bot = None
    aliases = None
    colors = True
    admin = None
    user_types = None
    rights = None

    def __init__(self, server, config):
        self.aliases = {}
        self.admin = True
        self.user_types = AttributeSet(['admin', 'irc'])
        self.rights = AttributeSet()
        for user_type in self.user_types:
            self.rights.update(commands.get_rights(user_type))
        self.server = server
        self.nickname = config.get('nickname',
                                   'piqueserver%s' % random.randrange(0, 99))
        self.username = config.get('username', 'piqueserver')
        self.realname = config.get('realname', server.name)
        self.channel = config.get('channel', "#piqueserver.bots").lower()
        self.commandprefix = config.get('commandprefix', '.')
        self.chatprefix = config.get('chatprefix', '')
        self.password = config.get('password', '') or None

    def startedConnecting(self, connector):
        log.info("Connecting to IRC server...")

    def clientConnectionLost(self, connector, reason):
        log.info("Lost connection to IRC server ({}), reconnecting in {} seconds".format(
            reason, self.lost_reconnect_delay))
        reactor.callLater(self.lost_reconnect_delay, connector.connect)

    def clientConnectionFailed(self, connector, reason):
        log.info("Could not connect to IRC server ({}), retrying in {} seconds".format(
            reason, self.failed_reconnect_delay))
        reactor.callLater(self.failed_reconnect_delay, connector.connect)

    def buildProtocol(self, address):
        p = self.protocol()
        p.factory = self
        p.protocol = self.server
        self.bot = p
        return p


class IRCRelay:
    factory = None

    def __init__(self, protocol, config):
        self.factory = IRCClientFactory(protocol, config)
        protocol.connectTCP(config.get('server'), config.get('port', 6667),
                            self.factory)

    def send(self, *arg, **kw):
        if self.factory.bot is None:
            return
        self.factory.bot.send(*arg, **kw)

    def me(self, *arg, **kw):
        if self.factory.bot is None:
            return
        self.factory.bot.me(*arg, **kw)


def format_name(player):
    return '{} #{}'.format(player.name, player.player_id)


def format_name_color(player):
    return (IRC_TEAM_COLORS.get(player.team.id, '') +
            '{} #{}'.format(player.name, player.player_id))


@restrict("irc")
@command()
def who(connection):
    protocol = connection.protocol
    player_count = len(protocol.players)
    if player_count == 0:
        connection.me('has no players connected')
        return
    sorted_players = sorted(protocol.players.values(),
                            key=attrgetter('team.id', 'name'))
    name_formatter = format_name_color if connection.colors else format_name
    teams = []
    formatted_names = []
    for k, g in groupby(sorted_players, attrgetter('team')):
        teams.append(k)
        formatted_names.append(map(name_formatter, g))
    separator = '\x0f, ' if connection.colors else ', '
    if not SPLIT_WHO_IN_TEAMS or player_count < SPLIT_THRESHOLD:
        noun = 'player' if player_count == 1 else 'players'
        msg = 'has {} {} connected: '.format(player_count, noun)
        msg += separator.join(chain.from_iterable(formatted_names))
        connection.me(msg)
    else:
        for team, names in zip(teams, formatted_names):
            name_count = len(names)
            noun = 'player' if name_count == 1 else 'players'
            msg = 'has {} {} in {}: '.format(name_count, noun, team.name)
            msg += separator.join(names)
            connection.me(msg)


@restrict("irc")
@command()
def score(connection):
    connection.me("scores: Blue {} - Green {}".format(
        connection.protocol.blue_team.score,
        connection.protocol.green_team.score))


@restrict("irc")
@command()
def alias(connection, value=None):
    aliases = connection.factory.aliases
    unaliased_name = connection.unaliased_name
    if value is None:
        alias = aliases.get(unaliased_name)
        if alias:
            message = 'aliases {} to {}'.format(unaliased_name, alias)
        else:
            message = "doesn't have an alias for %s" % unaliased_name
    else:
        aliases[unaliased_name] = value
        message = 'will alias {} to {}'.format(unaliased_name, value)
    connection.me(message)


@restrict("irc")
@command()
def unalias(connection):
    aliases = connection.factory.aliases
    unaliased_name = connection.unaliased_name
    if unaliased_name in aliases:
        aliases.pop(unaliased_name)
        message = 'will no longer alias %s' % unaliased_name
    else:
        message = "doesn't have an alias for %s" % unaliased_name
    connection.me(message)


@restrict("irc")
@command()
def colors(connection):
    connection.colors = not connection.colors
    if connection.colors:
        return '\x0312c\x0304o\x0309l\x0308o\x0306r\x0313s \x0f\x16ON!'
    else:
        return 'colors off'



"""
Protects spawned players for a specified amount of seconds.

Options
^^^^^^^

.. code-block:: guess

   [spawn_protect]
   protection_time = "3sec"

.. codeauthor:: ? & kmsi <kmsiapps@gmail.com>
"""

from pyspades.common import prettify_timespan
from piqueserver.config import config, cast_duration
from twisted.internet import reactor

spawn_protect_config = config.section("spawn_protect")
protection_time = spawn_protect_config.option("protection_time", default="3sec", cast=cast_duration)


def apply_script(protocol, connection, config):
    spawn_protect_time = protection_time.get()

    class SpawnProtectConnection(connection):
        spawn_timestamp = None

        def on_spawn(self, pos):
            self.spawn_timestamp = reactor.seconds()
            return connection.on_spawn(self, pos)

        def on_hit(self, hit_amount, hit_player, type, grenade):
            cur_timestamp = reactor.seconds() - spawn_protect_time
            if cur_timestamp < hit_player.spawn_timestamp:
                timespan = -(cur_timestamp - hit_player.spawn_timestamp)
                self.send_chat(
                    "%s is spawn-protected for %s." %
                    (hit_player.name, prettify_timespan(
                        timespan, True)))
                return False
            return connection.on_hit(
                self, hit_amount, hit_player, type, grenade)
    return protocol, SpawnProtectConnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import sys

from typing import List
import traceback

from twisted.internet import reactor
from twisted.protocols.basic import LineReceiver

from pyspades.types import AttributeSet
from piqueserver import commands

stdout = sys.__stdout__

if sys.platform == 'win32':
    # StandardIO on Windows does not work, so we create a silly replacement
    import msvcrt  # pylint: disable=import-error

    class StandardIO:
        disconnecting = False
        interval = 0.01
        input = ''

        def __init__(self, protocol):
            self.protocol = protocol
            protocol.makeConnection(self)
            self.get_input()

        def get_input(self):
            while msvcrt.kbhit():
                c = msvcrt.getwch()
                if c == '\r':  # new line
                    c = '\n'
                    stdout.write(c)
                    self.input += c
                    self.protocol.dataReceived(self.input)
                    self.input = ''
                elif c in ('\xE0', '\x00'):
                    # ignore special characters
                    msvcrt.getwch()
                elif c == '\x08':  # delete
                    self.input = self.input[:-1]
                    stdout.write('\x08 \x08')
                else:
                    self.input += c
                    stdout.write(c)
            reactor.callLater(self.interval, self.get_input)

        def write(self, data):
            stdout.write(data)

        def writeSequence(self, seq):
            stdout.write(''.join(seq))
else:
    from twisted.internet.stdio import StandardIO


class ConsoleInput(LineReceiver):
    name = 'Console'
    admin = True
    delimiter = b'\n'

    def __init__(self, protocol):
        self.protocol = protocol
        self.user_types = AttributeSet(['admin', 'console'])
        self.rights = AttributeSet()
        for user_type in self.user_types:
            self.rights.update(commands.get_rights(user_type))

    def lineReceived(self, line):
        if not line:
            return

        try:
            result = commands.handle_input(self, line.decode())
        # pylint: disable=broad-except
        except Exception:
            traceback.print_exc()
        else:
            if result is not None:
                print(result)

    # methods used to emulate the behaviour of regular Connection objects to
    # prevent errors when command writers didn't test that their scripts would
    # work when run on the console
    def send_chat(self, value: str, _):
        print(value)

    def send_lines(self, lines: List[str]):
        print("\n".join(lines))


def create_console(protocol):
    console = ConsoleInput(protocol)
    StandardIO(console)



# Copyright (c) junk/someonesomewhere 2011.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from twisted.internet.defer import Deferred
import aiohttp
from aiohttp import web
from multidict import MultiDict

from jinja2 import Environment, PackageLoader
import json
import time
from PIL import Image
from io import BytesIO
from aiohttp.abc import AbstractAccessLogger
from twisted.logger import Logger
from piqueserver.utils import as_deferred

from piqueserver.config import config, cast_duration

status_server_config = config.section("status_server")
host_option = status_server_config.option("host", "0.0.0.0")
port_option = status_server_config.option("port", 32886)
logging_option = status_server_config.option("logging", False)
interval_option = status_server_config.option(
    "update_interval", default="1min", cast=cast_duration)
scripts_option = config.option("scripts", [])


class AccessLogger(AbstractAccessLogger):

    def log(self, request, response, time):
        self.logger.info(
            "{remote} {method} {url}: {status} {time:0.2f}ms -- {ua}",
            remote=request.remote,
            ua=request.headers["User-Agent"],
            method=request.method,
            url=request.url,
            time=time * 1000,
            status=response.status)


async def set_default_headers(request, response):
    response.headers['Access-Control-Allow-Origin'] = '*'
    response.headers['Access-Control-Allow-Credentials'] = 'true'


def current_state(protocol):
    """Gathers data on current server/game state from protocol class"""
    players = []

    for player in protocol.players.values():
        player_data = {
            'name': player.name,
            'latency': player.latency,
            'client': player.client_string,
            'kills': player.kills,
            'team': player.team.name
        }

        players.append(player_data)

    dictionary = {
        "serverIdentifier": protocol.identifier,
        "serverName": protocol.name,
        "serverVersion": protocol.version,
        "serverUptime": time.time() - protocol.start_time,
        "gameMode": protocol.game_mode_name,
        "map": {
            "name": protocol.map_info.name,
            "version": protocol.map_info.version,
            "author": protocol.map_info.author
        },
        "scripts": scripts_option.get(),
        "players": players,
        "maxPlayers": protocol.max_players,
        "scores": {
            "currentBlueScore": protocol.blue_team.score,
            "currentGreenScore": protocol.green_team.score,
            "maxScore": protocol.max_score}
    }

    return dictionary


class StatusServer:
    def __init__(self, protocol):
        self.protocol = protocol
        self.last_update = None
        self.last_map_name = None
        self.cached_overview = None
        env = Environment(loader=PackageLoader('piqueserver.web'))
        self.status_template = env.get_template('status.html')

    async def json(self, request):
        state = current_state(self.protocol)
        return web.json_response(state)

    @property
    def current_map(self):
        return self.protocol.map_info.name

    def update_cached_overview(self):
        """Updates cached overview"""
        overview = self.protocol.map.get_overview(rgba=True)
        image = Image.frombytes('RGBA', (512, 512), overview)
        data = BytesIO()
        image.save(data, 'png')
        self.cached_overview = data.getvalue()
        self.last_update = time.time()
        self.last_map_name = self.protocol.map_info.name

    async def overview(self, request):
        # update cache on a set interval or map change or initialization
        if (self.cached_overview is None or
                self.last_map_name != self.current_map or
                time.time() - self.last_update > interval_option.get()):
            self.update_cached_overview()

        return web.Response(body=self.cached_overview,
                            content_type='image/png')

    async def index(self, request):
        rendered = self.status_template.render(server=self.protocol)
        return web.Response(body=rendered, content_type='text/html')

    def create_app(self):
        app = web.Application()
        app.on_response_prepare.append(set_default_headers)
        app.add_routes([
            web.get('/json', self.json),
            web.get('/overview', self.overview),
            web.get('/', self.index)
        ])
        return app

    async def listen(self):
        """Starts the status server on configured host/port"""
        app = self.create_app()
        logger = Logger() if logging_option.get() else None
        log_class = AccessLogger if logging_option.get() else None
        runner = web.AppRunner(app,
                               access_log=logger,
                               access_log_class=log_class)
        await as_deferred(runner.setup())
        site = web.TCPSite(runner, host_option.get(), port_option.get())
        await as_deferred(site.start())

        # TODO: explain why we do this
        await Deferred()



"""
Rollback rolls back the map to it's original state by placing and removing
changed blocks. This takes ages. Use with care.

Commands
^^^^^^^^

* ``/rollmap <map name>`` changes the map to the given map in a rolling fashion *admin only*
* ``/rollmap <map name> <sector>`` changes the sector of the current map to the given map's sector in a rolling fashion *admin only*
* ``/rollback`` starts a rollback on the current map *admin only*
* ``/rollbackcancel`` cancel an on-going rollback *admin only*

.. warning::
   ``/rollmap`` will take a long time if number of differing blocks is too high.

Options
^^^^^^^

.. code-block:: guess

    [rollback]
    rollback_on_game_end = false

.. codeauthor:: hompy
"""

import os
import time
import operator

from twisted.internet.task import LoopingCall
from pyspades.vxl import VXLData
from pyspades.contained import BlockAction, SetColor
from pyspades.constants import *
from pyspades.common import coordinates, make_color
from piqueserver.map import Map, MapNotFound, check_rotation
from piqueserver.commands import command, admin
from piqueserver.config import config

S_INVALID_MAP_NAME = 'Invalid map name'
S_ROLLBACK_IN_PROGRESS = 'Rollback in progress'
S_ROLLBACK_COMMENCED = '{player} commenced a rollback...'
S_AUTOMATIC_ROLLBACK_PLAYER_NAME = 'Map'
S_NO_ROLLBACK_IN_PROGRESS = 'No rollback in progress'
S_ROLLBACK_CANCELLED = 'Rollback cancelled by {player}'
S_ROLLBACK_ENDED = 'Rollback ended. {result}'
S_MAP_CHANGED = 'Map was changed'
S_ROLLBACK_PROGRESS = 'Rollback progress {percent:.0%}'
S_ROLLBACK_COLOR_PASS = 'Rollback doing color pass...'
S_ROLLBACK_TIME_TAKEN = 'Time taken: {seconds:.3}s'

NON_SURFACE_COLOR = (0, 0, 0)

rollback_config = config.section('rollback')
ROLLBACK_ON_GAME_END_OPTION = rollback_config.option('rollback_on_game_end', False)
config_dir = config.config_dir

@command(admin_only=True)
def rollmap(connection, mapname=None, value=None):
    start_x, start_y, end_x, end_y = 0, 0, 512, 512
    if value is not None:
        start_x, start_y = coordinates(value)
        end_x, end_y = start_x + 64, start_y + 64
    return connection.protocol.start_rollback(connection, mapname,
                                              start_x, start_y, end_x, end_y)


@command(admin_only=True)
def rollback(connection, value=None):
    return rollmap(connection, value=value)


@command(admin_only=True)
def rollbackcancel(connection):
    return connection.protocol.cancel_rollback(connection)


def apply_script(protocol, connection, config):
    rollback_on_game_end = ROLLBACK_ON_GAME_END_OPTION.get()

    class RollbackConnection(connection):

        def on_block_destroy(self, x, y, z, value):
            if self.protocol.rollback_in_progress:
                return False
            return connection.on_block_destroy(self, x, y, z, value)

    class RollbackProtocol(protocol):
        rollback_in_progress = False
        rollback_max_rows = 10  # per 'cycle', intended to cap cpu usage
        # per 'cycle' cap for (unique packets * players)
        rollback_max_packets = 180
        rollback_max_unique_packets = 12  # per 'cycle', each block op is at least 1
        rollback_time_between_cycles = 0.06
        rollback_time_between_progress_updates = 10.0
        rollback_start_time = None
        rollback_last_chat = None
        rollback_rows = None
        rollback_total_rows = None

        # rollback

        def start_rollback(self, connection, mapname, start_x, start_y,
                           end_x, end_y, ignore_indestructable=True):
            if self.rollback_in_progress:
                return S_ROLLBACK_IN_PROGRESS
            if mapname is None:
                map = self.rollback_map
            else:
                try:
                    maps = check_rotation([mapname])
                    if not maps:
                        return S_INVALID_MAP_NAME
                    map = Map(maps[0], os.path.join(config_dir, "maps")).data
                except MapNotFound as error:
                    return error.message
            name = (connection.name if connection is not None
                    else S_AUTOMATIC_ROLLBACK_PLAYER_NAME)
            message = S_ROLLBACK_COMMENCED.format(player=name)
            self.send_chat(message, irc=True)
            self.packet_generator = self.create_rollback_generator(
                self.map, map, start_x, start_y, end_x, end_y, ignore_indestructable)
            self.rollback_in_progress = True
            self.rollback_start_time = time.monotonic()
            self.rollback_last_chat = self.rollback_start_time
            self.rollback_rows = 0
            self.rollback_total_rows = end_x - start_x
            self.cycle_call = LoopingCall(self.rollback_cycle)
            self.cycle_call.start(self.rollback_time_between_cycles)

        def cancel_rollback(self, connection):
            if not self.rollback_in_progress:
                return S_NO_ROLLBACK_IN_PROGRESS
            result = S_ROLLBACK_CANCELLED.format(player=connection.name)
            self.end_rollback(result)

        def end_rollback(self, result):
            self.rollback_in_progress = False
            self.cycle_call.stop()
            self.cycle_call = None
            self.packet_generator = None
            self.update_entities()
            message = S_ROLLBACK_ENDED.format(result=result)
            self.send_chat(message, irc=True)

        def rollback_cycle(self):
            if not self.rollback_in_progress:
                return
            try:
                sent_unique = sent_total = rows = 0
                while True:
                    if rows > self.rollback_max_rows:
                        break
                    if sent_unique > self.rollback_max_unique_packets:
                        break
                    if sent_total > self.rollback_max_packets:
                        break
                    sent = next(self.packet_generator)
                    sent_unique += sent
                    sent_total += sent * len(self.connections)
                    rows += (sent == 0)
                self.rollback_rows += rows
                if (time.monotonic() - self.rollback_last_chat >
                        self.rollback_time_between_progress_updates):
                    self.rollback_last_chat = time.monotonic()
                    progress = float(self.rollback_rows) / \
                        self.rollback_total_rows
                    if progress < 1.0:
                        message = S_ROLLBACK_PROGRESS.format(percent=progress)
                        self.send_chat(message)
                    else:
                        self.send_chat(S_ROLLBACK_COLOR_PASS)
            except (StopIteration):
                elapsed = time.monotonic() - self.rollback_start_time
                message = S_ROLLBACK_TIME_TAKEN.format(seconds=elapsed)
                self.end_rollback(message)

        def create_rollback_generator(self, cur, new, start_x, start_y,
                                      end_x, end_y, ignore_indestructable):
            surface = {}
            block_action = BlockAction()
            block_action.player_id = 31
            set_color = SetColor()
            set_color.value = make_color(*NON_SURFACE_COLOR)
            set_color.player_id = 31
            self.send_contained(set_color, save=True)
            old = cur.copy()
            check_protected = hasattr(protocol, 'protected')
            for x in range(start_x, end_x):
                block_action.x = x
                for y in range(start_y, end_y):
                    block_action.y = y
                    if check_protected and self.is_protected(x, y, 0):
                        continue
                    for z in range(63):
                        action = None
                        cur_solid = cur.get_solid(x, y, z)
                        new_solid = new.get_solid(x, y, z)
                        if cur_solid and not new_solid:
                            if (not ignore_indestructable and
                                    self.is_indestructable(x, y, z)):
                                continue
                            else:
                                action = DESTROY_BLOCK
                                cur.remove_point(x, y, z)
                        elif new_solid:
                            new_is_surface = new.is_surface(x, y, z)
                            if new_is_surface:
                                new_color = new.get_color(x, y, z)
                            if not cur_solid and new_is_surface:
                                surface[(x, y, z)] = new_color
                            elif not cur_solid and not new_is_surface:
                                action = BUILD_BLOCK
                                cur.set_point(x, y, z, NON_SURFACE_COLOR)
                            elif cur_solid and new_is_surface:
                                old_is_surface = old.is_surface(x, y, z)
                                if old_is_surface:
                                    old_color = old.get_color(x, y, z)
                                if not old_is_surface or old_color != new_color:
                                    surface[(x, y, z)] = new_color
                                    action = DESTROY_BLOCK
                                    cur.remove_point(x, y, z)
                        if action is not None:
                            block_action.z = z
                            block_action.value = action
                            self.send_contained(block_action, save=True)
                            yield 1
                yield 0
            last_color = None
            block_action.value = BUILD_BLOCK
            for pos, color in sorted(iter(surface.items()),
                                     key=operator.itemgetter(1)):
                x, y, z = pos
                packets_sent = 0
                if color != last_color:
                    set_color.value = make_color(*color)
                    self.send_contained(set_color, save=True)
                    packets_sent += 1
                    last_color = color
                cur.set_point(x, y, z, color)
                block_action.x = x
                block_action.y = y
                block_action.z = z
                self.send_contained(block_action, save=True)
                packets_sent += 1
                yield packets_sent

        def on_map_change(self, map):
            self.rollback_map = map.copy()
            protocol.on_map_change(self, map)

        def on_map_leave(self):
            if self.rollback_in_progress:
                self.end_rollback(S_MAP_CHANGED)
            protocol.on_map_leave(self)

        def on_game_end(self):
            if rollback_on_game_end:
                self.start_rollback(None, None, 0, 0, 512, 512, False)
            protocol.on_game_end(self)

    return RollbackProtocol, RollbackConnection



"""
The map generator is responsible for generating the map bytes that get sent
to the client on connect
"""
import zlib

COMPRESSION_LEVEL = 9


class ProgressiveMapGenerator:
    """
    Progressively generates the stream of bytes sent to the client for map
    downloads.

    It supports two modes. In the default parent=False mode, reading is normal.

    In the `parent=True` mode a child generator is created with `get_child` to
    actually read the data. This is presumably done so that the work of map
    generation is not duplicated for each client if several connect at the same
    time.
    """
    data = b''
    done = False

    # parent attributes
    all_data = b''
    pos = 0

    def __init__(self, map_, parent=False):
        # parent=True enables saving all data sent instead of just
        # deleting it afterwards.
        self.parent = parent
        self.generator = map_.get_generator()
        self.compressor = zlib.compressobj(COMPRESSION_LEVEL)

    def get_size(self):
        """get the map size, for display of the loading bar on the client"""
        # This is currently just an estimate, since due to compression
        # magic, we don't actually know what size the file will be when sent
        # over the wire
        return 1.5 * 1024 * 1024  # 2MB

    def read(self, size):
        """read size bytes from the map generator"""
        data = self.data
        generator = self.generator
        if len(data) < size and generator is not None:
            while True:
                map_data = generator.get_data(size)
                if generator.done:
                    self.generator = None
                    data += self.compressor.flush()
                    break
                data += self.compressor.compress(map_data)
                if len(data) >= size:
                    break
        if self.parent:
            # save the data in case we are a parent
            self.all_data += data
            self.pos += len(data)
        else:
            self.data = data[size:]
            return data[:size]

    def get_child(self):
        """return a new child generator"""
        if self.parent:
            return MapGeneratorChild(self)
        else:
            raise NotImplementedError(
                "get_child is not implemented for non-parent generators")

    def data_left(self):
        """return True if any data is left"""
        return bool(self.data) or self.generator is not None


class MapGeneratorChild:
    pos = 0

    def __init__(self, generator):
        self.parent = generator

    def get_size(self):
        """get the size of the parent map generator"""
        return self.parent.get_size()

    def read(self, size):
        """read size bytes from the parent map generator, if possible"""
        pos = self.pos
        if pos + size > self.parent.pos:
            self.parent.read(size)
        data = self.parent.all_data[pos:pos + size]
        self.pos += len(data)
        return data

    def data_left(self):
        """return True if any data is left"""
        return self.parent.data_left() or self.pos < self.parent.pos



from random import choice
from twisted.internet import reactor
# aparently, we need to send packets in this file. For now, I give in.
from pyspades.contained import CreatePlayer, SetTool, KillAction, InputData, SetColor, WeaponInput
from pyspades.constants import (GRENADE_KILL, FALL_KILL, NETWORK_FPS)
from pyspades.common import (
    prettify_timespan,
    make_color)
from piqueserver.commands import (
    command, CommandError, get_player, join_arguments, target_player)
from piqueserver.utils import timeparse


def has_digits(s: str) -> bool:
    return any(char.isdigit() for char in s)


def get_ban_arguments(connection, args):
    """
    Parses duration and reason from arguments.
    It handles duration in two ways: interger mintues and human-friendly duration.
    It also handles cases where duration or reason are none.
    Note: It returns duration in seconds.
    """
    default_duration = connection.protocol.default_ban_time
    reason = None
    if len(args) < 1:
        return default_duration, reason
    if len(args) > 1:
        reason = join_arguments(args[1:])
    if args[0] == "perma":
        return None, reason

    if args[0].isdigit():  # all digits == duration in minutes
        duration = int(args[0]) * 60
    elif has_digits(args[0]):  # if it contains some digits maybe duration?
        duration = timeparse(args[0])
        if not duration:
            raise ValueError("Invalid duration")
    else:  # maybe just one long reason
        duration = default_duration
        reason = join_arguments(args[:])
    return duration, reason


@command(admin_only=True)
def kick(connection, value, *arg):
    """
    Kick a given player
    /kick <player>
    Player is the #ID of the player, or an unique part of their name
    """
    reason = join_arguments(arg)
    player = get_player(connection.protocol, value)
    player.kick(reason)


@command(admin_only=True)
def ban(connection, value, *arg):
    """
    Ban a given player forever or for a limited amount of time.
    /ban <player> [duration] [reason]
    """
    duration, reason = get_ban_arguments(connection, arg)
    player = get_player(connection.protocol, value)
    player.ban(reason, duration)


@command(admin_only=True)
def hban(connection, value, *arg):
    """
    Ban a given player for an hour
    /hban <player> [reason]
    """
    duration = timeparse("1hour")
    reason = join_arguments(arg)
    player = get_player(connection.protocol, value)
    player.ban(reason, duration)


@command(admin_only=True)
def dban(connection, value, *arg):
    """
    Ban a given player for one day
    /dban <player> [reason]
    """
    duration = timeparse("1day")
    reason = join_arguments(arg)
    player = get_player(connection.protocol, value)
    player.ban(reason, duration)


@command(admin_only=True)
def wban(connection, value, *arg):
    """
    Ban a given player for one week
    /wban <player> [reason]
    """
    duration = timeparse("1week")
    reason = join_arguments(arg)
    player = get_player(connection.protocol, value)
    player.ban(reason, duration)


@command(admin_only=True)
def pban(connection, value, *arg):
    """
    Ban a given player permanently
    /pban <player> [reason]
    """
    duration = 0
    reason = join_arguments(arg)
    player = get_player(connection.protocol, value)
    player.ban(reason, duration)


@command(admin_only=True)
def banip(connection, ip, *arg):
    """
    Ban an ip
    /banip <ip> [duration] [reason]
    """
    duration, reason = get_ban_arguments(connection, arg)
    try:
        connection.protocol.add_ban(ip, reason, duration)
    except ValueError:
        return 'Invalid IP address/network'
    reason = ': ' + reason if reason is not None else ''
    duration = duration or None
    if duration is None:
        return 'IP/network %s permabanned%s' % (ip, reason)
    else:
        return 'IP/network %s banned for %s%s' % (
            ip, prettify_timespan(duration), reason)


@command(admin_only=True)
def unban(connection, ip):
    """
    Unban an ip
    /unban <ip>
    """
    try:
        connection.protocol.remove_ban(ip)
        return 'IP unbanned'
    except KeyError:
        return 'IP not found in ban list'


@command('undoban', admin_only=True)
def undo_ban(connection, *arg):
    """
    Undo last ban
    /undoban
    """
    if len(connection.protocol.bans) > 0:
        result = connection.protocol.undo_last_ban()
        return 'Ban for %s undone' % result[0]
    else:
        return 'No bans to undo!'


@command(admin_only=True)
def say(connection, *arg):
    """
    Say something in chat as server message
    /say <text>
    """
    value = ' '.join(arg)
    connection.protocol.send_chat(value)
    connection.protocol.irc_say(value)


@command(admin_only=True)
def mute(connection, value):
    """
    Mute a player
    /mute <player>
    """
    player = get_player(connection.protocol, value)
    if player.mute:
        return '%s is already muted' % player.name
    player.mute = True
    message = '%s has been muted by %s' % (player.name, connection.name)
    connection.protocol.send_chat(message, irc=True)


@command(admin_only=True)
def unmute(connection, value):
    """
    Unmute a player
    /unmute <player>
    """
    player = get_player(connection.protocol, value)
    if not player.mute:
        return '%s is not muted' % player.name
    player.mute = False
    message = '%s has been unmuted by %s' % (player.name, connection.name)
    connection.protocol.send_chat(message, irc=True)


@command(admin_only=True)
@target_player
def ip(connection, player):
    """
    Get the IP of a user
    /ip [player]
    """
    return 'The IP of %s is %s' % (player.name, player.address[0])


@command("whowas", admin_only=True)
def who_was(connection, value):
    """
    Get the IP of a user who has left
    /whowas <player>
    """
    value = value.lower()
    ret = None
    exact_match = False
    for name, ip in connection.protocol.player_memory:
        name_lower = name.lower()
        if name_lower == value:
            ret = (name, ip)
            exact_match = True
        elif not exact_match and name_lower.count(value):
            ret = (name, ip)
    if ret is None:
        raise CommandError("Invalid Player")
    return "%s's most recent IP was %s" % ret


@command('invisible', 'invis', 'inv', admin_only=True)
@target_player
def invisible(connection, player):
    """
    Turn invisible
    /invisible [player]
    """
    protocol = connection.protocol
    # TODO: move this logic to a more suitable place
    player.invisible = not player.invisible
    player.filter_visibility_data = player.invisible
    player.god = player.invisible
    player.god_build = False
    player.killing = not player.invisible
    if player.invisible:
        player.send_chat("You're now invisible")
        protocol.irc_say('* %s became invisible' % player.name)
        kill_action = KillAction()
        kill_action.kill_type = choice([GRENADE_KILL, FALL_KILL])
        kill_action.player_id = kill_action.killer_id = player.player_id
        reactor.callLater(1.0 / NETWORK_FPS, protocol.broadcast_contained,
                          kill_action, sender=player)
    else:
        player.send_chat("You return to visibility")
        protocol.irc_say('* %s became visible' % player.name)
        x, y, z = player.world_object.position.get()
        create_player = CreatePlayer()
        create_player.player_id = player.player_id
        create_player.name = player.name
        create_player.x = x
        create_player.y = y
        create_player.z = z
        create_player.weapon = player.weapon
        create_player.team = player.team.id
        world_object = player.world_object
        input_data = InputData()
        input_data.player_id = player.player_id
        input_data.up = world_object.up
        input_data.down = world_object.down
        input_data.left = world_object.left
        input_data.right = world_object.right
        input_data.jump = world_object.jump
        input_data.crouch = world_object.crouch
        input_data.sneak = world_object.sneak
        input_data.sprint = world_object.sprint
        set_tool = SetTool()
        set_tool.player_id = player.player_id
        set_tool.value = player.tool
        set_color = SetColor()
        set_color.player_id = player.player_id
        set_color.value = make_color(*player.color)
        weapon_input = WeaponInput()
        weapon_input.primary = world_object.primary_fire
        weapon_input.secondary = world_object.secondary_fire
        protocol.broadcast_contained(create_player, sender=player, save=True)
        protocol.broadcast_contained(set_tool, sender=player)
        protocol.broadcast_contained(set_color, sender=player, save=True)
        protocol.broadcast_contained(input_data, sender=player)
        protocol.broadcast_contained(weapon_input, sender=player)
    if connection is not player and connection in protocol.players.values():
        if player.invisible:
            return '%s is now invisible' % player.name
        else:
            return '%s is now visible' % player.name


@command('godsilent', 'gods', admin_only=True)
def godsilent(connection, player=None):
    """
    Silently go into god mode
    /godsilent [player]
    """
    if player is not None:
        connection = get_player(connection.protocol, player)
    elif connection not in connection.protocol.players.values():
        return 'Unknown player: ' + player

    connection.god = not connection.god  # toggle godmode

    if connection.protocol.set_god_build:
        connection.god_build = connection.god
    else:
        connection.god_build = False

    if connection.god:
        if player is None:
            return 'You have silently entered god mode'
        else:
            # TODO: Do not send this if the specified player is the one who called the command
            connection.send_chat(
                'Someone has made you silently enter godmode!')
            return 'You made ' + connection.name + ' silently enter god mode'
    else:
        if player is None:
            return 'You have silently returned to being a mere human'
        else:
            # TODO: Do not send this if the specified player is the one who called the command
            connection.send_chat(
                'Someone has made you silently return to being a mere human')
            return 'You made ' + connection.name + ' silently return to being a mere human'


@command(admin_only=True)
def god(connection, player=None):
    """
    Go into god mode and inform everyone on the server of it
    /god [player]
    """
    if player:
        connection = get_player(connection.protocol, player)
    elif connection not in connection.protocol.players.values():
        return 'Unknown player'

    connection.god = not connection.god  # toggle godmode

    if connection.god:
        message = '%s entered GOD MODE!' % connection.name
    else:
        message = '%s returned to being a mere human' % connection.name
    connection.protocol.send_chat(message, irc=True)


@command('godbuild', admin_only=True)
@target_player
def god_build(connection, player):
    """
    Place blocks that can be destroyed only by players with godmode activated
    /godbuild [player]
    """
    protocol = connection.protocol

    player.god_build = not player.god_build

    message = ('now placing god blocks' if player.god_build else
               'no longer placing god blocks')
    player.send_chat("You're %s" % message)
    if connection is not player and connection in protocol.players.values():
        connection.send_chat('%s is %s' % (player.name, message))
    protocol.irc_say('* %s is %s' % (player.name, message))



import abc
from typing import Tuple


class BaseAuthBackend(abc.ABC):
    @abc.abstractmethod
    def login(self, details: Tuple):
        pass

class ConfigAuthBackend(BaseAuthBackend):
    """Auth backend that uses the [passwords] section of the connfig for
    authentication"""
    def login(self, username):
        pass



from ipaddress import ip_network, ip_address
from collections import OrderedDict

def get_cidr(network):
    if network.prefixlen == 32:
        return str(network.network_address)
    return str(network)

# Note: Network objects cannot have any host bits set without strict=False.
# More info: https://docs.python.org/3/howto/ipaddress.html#defining-networks


class NetworkDict:
    def __init__(self):
        self.networks = OrderedDict()

    def read_list(self, values):
        for index, item in enumerate(values):
            if len(item) < 4:
                raise ValueError("Invalid ban entry. index: {} item: {}\nEntry format needs to be [name, ip, reason, time]".format(index, item))
            self[item[1]] = [item[0]] + item[2:]

    def make_list(self):
        values = []
        for network, value in self.iteritems():
            values.append([value[0]] + [network] + list(value[1:]))
        return values

    def remove(self, key):
        """remove a key from the networkdict and return the removed items"""
        ip = ip_network(str(key), strict=False)
        results = []
        # There are 32 possible networks for each IP address in CIDR. This is
        # small enough that we can just loop through all of them to get an
        # unelegant constant time lookup for IP addresses.
        #
        # This loop could be sped up by a lot by using a (ip, mask) int tuple
        # instead of constantly creating new IPNetwork objects with .supernet()
        #
        # When in doubt, use brute force - Ken Thompson
        for _ in range(0, 32):
            try:
                elem = self.networks.pop(ip)
            except KeyError:
                pass
            else:
                results.append([ip, elem])
            ip = ip.supernet()
        return results

    def __setitem__(self, key, value):
        self.networks[ip_network(str(key), strict=False)] = value

    def __getitem__(self, key):
        return self.get_entry(key)[1]

    def get_entry(self, key):
        ip = ip_network(str(key))
        # See comment for remove()
        for _ in range(0, 32):
            try:
                return self.networks[ip]
            except KeyError:
                pass
            ip = ip.supernet()
        raise KeyError(key)

    def __len__(self):
        return len(self.networks)

    def __delitem__(self, key):
        ip = ip_network(str(key), strict=False)
        self.networks.popitem(ip)

    def pop(self, *arg, **kw):
        network, value = self.networks.popitem(*arg, **kw)
        return get_cidr(network), value

    def iteritems(self):
        for network, value in self.networks.items():
            yield get_cidr(network), value

    def __contains__(self, key):
        try:
            self.get_entry(key)
            return True
        except KeyError:
            return False



"""
Adds the ability to 'trust' certain players, i.e. they cannot be votekicked
or rubberbanded.

Commands
^^^^^^^^

* ``/trust <player>`` gives trusted status to a player

.. codeauthor:: mat^2 & hompy
"""

from piqueserver.commands import command, admin, get_player

S_GRANTED = '{player} is now trusted'
S_GRANTED_SELF = "You've been granted trust, and can't be votekicked"
S_CANT_VOTEKICK = "{player} is trusted and can't be votekicked"
S_RESULT_TRUSTED = 'Trusted user'


@command(admin_only=True)
def trust(connection, player):
    player = get_player(connection.protocol, player)
    player.on_user_login('trusted', False)
    player.send_chat(S_GRANTED_SELF)
    return S_GRANTED.format(player=player.name)


def apply_script(protocol, connection, config):
    class TrustedConnection(connection):

        def on_user_login(self, user_type, verbose=True):
            if user_type == 'trusted':
                self.speedhack_detect = False
                votekick = getattr(self.protocol, 'votekick', None)
                if votekick and votekick.victim is self:
                    votekick.end(S_RESULT_TRUSTED)
                    self.protocol.votekick = None
            return connection.on_user_login(self, user_type, verbose)

    class TrustedProtocol(protocol):

        def on_votekick_start(self, instigator, victim, reason):
            if victim.user_types.trusted:
                return S_CANT_VOTEKICK.format(player=victim.name)
            return protocol.on_votekick_start(self, instigator, victim, reason)

    return TrustedProtocol, TrustedConnection



"""
Team Deathmatch game mode.

Maintainer: Triplefox
"""

from pyspades.constants import *

from piqueserver.commands import command

HIDE_COORD = (0, 0, 0)


@command()
def score(connection):
    return connection.protocol.get_kill_count()


def apply_script(protocol, connection, config):
    class TDMConnection(connection):

        def on_spawn(self, pos):
            self.send_chat(self.explain_game_mode())
            self.send_chat(self.protocol.get_kill_count())
            return connection.on_spawn(self, pos)

        def on_flag_take(self):
            if self.protocol.remove_intel:
                return False
            return connection.on_flag_take(self)

        def on_flag_capture(self):
            result = connection.on_flag_capture(self)
            self.team.kills += self.protocol.intel_points
            self.protocol.check_end_game(self)
            return result

        def on_kill(self, killer, type, grenade):
            result = connection.on_kill(self, killer, type, grenade)
            self.protocol.check_end_game(killer)
            return result

        def explain_game_mode(self):
            msg = 'Team Deathmatch: Kill the opposing team.'
            if not self.protocol.remove_intel:
                msg += ' Intel is worth %s kills.' % self.protocol.intel_points
            return msg

    class TDMProtocol(protocol):
        game_mode = CTF_MODE
        kill_limit = config.get('kill_limit', 100)
        intel_points = config.get('intel_points', 10)
        remove_intel = config.get('remove_intel', False)

        def on_flag_spawn(self, x, y, z, flag, entity_id):
            if self.remove_intel:
                return HIDE_COORD
            return protocol.on_flag_spawn(self, x, y, z, flag, entity_id)

        def get_kill_count(self):
            green_kills = self.green_team.kills
            blue_kills = self.blue_team.kills
            diff = green_kills - blue_kills
            if green_kills > blue_kills:
                return ("Green leads %s-%s (+%s, %s left). Playing to %s kills." %
                        (green_kills, blue_kills,
                         diff,
                         self.kill_limit - green_kills,
                         self.kill_limit))
            elif green_kills < blue_kills:
                return ("Blue leads %s-%s (+%s, %s left). Playing to %s kills." %
                        (blue_kills, green_kills,
                         -diff,
                         self.kill_limit - blue_kills,
                         self.kill_limit))
            else:
                return ("%s-%s, %s left. Playing to %s kills." %
                        (green_kills,
                         blue_kills,
                         self.kill_limit - green_kills,
                         self.kill_limit))

        def check_end_game(self, player):
            if self.green_team.kills >= self.kill_limit:
                self.send_chat("Green Team Wins, %s - %s" %
                               (self.green_team.kills, self.blue_team.kills))
                self.reset_game(player)
                protocol.on_game_end(self)
            elif self.blue_team.kills >= self.kill_limit:
                self.send_chat("Blue Team Wins, %s - %s" %
                               (self.blue_team.kills, self.green_team.kills))
                self.reset_game(player)
                protocol.on_game_end(self)

    return TDMProtocol, TDMConnection



"""
Shows a detailed analysis of a players shots

Commands
^^^^^^^^

* ``/analyze or /an <target>`` to show a detailed analysis of a players shots if they hit someone.
  Shows hit player, distance, dT in milliseconds (Delta Time- Time since previous shot that hit someone.
  Useful for detecting multiple bullet or rapid hacks), weapon, which body part it hit, and a basic counter that displays the number of hits of that type.

* ``/analyze or /an`` to disable it

.. codeauthor:: a_girl
 """

import time
from pyspades.collision import distance_3d_vector
from pyspades.constants import (HEADSHOT_KILL)
from piqueserver.commands import command, get_player


@command('analyze', 'an')
def analyze_shot(connection, player=None):
    protocol = connection.protocol
    if not protocol.analyzers:
        protocol.analyzers = {}
    if player is None:
        if connection.name in protocol.analyzers:
            del protocol.analyzers[connection.name]
            return 'You are no longer analyzing anyone.'

        return 'Please enter a target player to analyze.'

    player = get_player(protocol, player)
    if player not in protocol.players:
        raise ValueError("Target player is required")

    if connection.name in protocol.analyzers:
        if player.name == protocol.analyzers[connection.name]:
            del protocol.analyzers[connection.name]
            return 'You are no longer analyzing anyone.'

        protocol.analyzers[connection.name] = player.name
        connection.hs, connection.bs, connection.ls = 0, 0, 0
        return (
            'You are no longer analyzing {}. You are now analyzing {}.'.format(
                protocol.analyzers[connection.name], player.name))

    protocol.analyzers[connection.name] = player.name
    connection.hs, connection.bs, connection.ls = 0, 0, 0
    return 'You are now analyzing {}'.format(player.name)

body_damage_values = [49, 29, 27]
limb_damage_values = [33, 18, 16]

def apply_script(protocol, connection, config):
    class analyze_shotsConnection(connection):
        dist = ""
        weap = ""
        hs, bs, ls = 0, 0, 0
        prev_time = None
        body_part = ""

        def on_hit(self, hit_amount, hit_player, type, grenade):
            if self.name in list(self.protocol.analyzers.values()):
                if type == HEADSHOT_KILL or hit_amount in body_damage_values or hit_amount in limb_damage_values:
                    if not grenade:
                        dist = int(distance_3d_vector(
                            self.world_object.position, hit_player.world_object.position))
                        weap = self.weapon_object.name
                        self.pres_time = time.monotonic()
                        if self.prev_time is None:
                            dt = None
                        else:
                            dt = (self.pres_time - self.prev_time) * 1000
                        self.prev_time = time.monotonic()
                        if type == HEADSHOT_KILL:
                            body_part = "HEADSHOT"
                        elif hit_amount in body_damage_values:
                            body_part = "Body"
                        elif hit_amount in limb_damage_values:
                            body_part = "Limb"
                        for name in list(self.protocol.analyzers.keys()):
                            if self.protocol.analyzers[name] == self.name:
                                analyzer = get_player(self.protocol, name)
                                if analyzer not in self.protocol.players.values():
                                    raise ValueError()
                                else:
                                    if body_part == "HEADSHOT":
                                        analyzer.hs += 1
                                        counter = analyzer.hs
                                    elif body_part == "Body":
                                        analyzer.bs += 1
                                        counter = analyzer.bs
                                    elif body_part == "Limb":
                                        analyzer.ls += 1
                                        counter = analyzer.ls
                                    if dt is not None:
                                        analyzer.send_chat('%s shot %s dist: %d blocks dT: %.0f ms %s %s(%d)' % (
                                            self.name, hit_player.name, dist, dt, weap, body_part, counter))
                                    else:
                                        analyzer.send_chat('%s shot %s dist: %d blocks dT: NA %s %s(%d)' % (
                                            self.name, hit_player.name, dist, weap, body_part, counter))
            return connection.on_hit(self, hit_amount, hit_player, type, grenade)

        def on_weapon_set(self, value):
            if self.name in list(self.protocol.analyzers.values()):
                for name in self.protocol.analyzers:
                    if self.protocol.analyzers[name] == self.name:
                        analyzer = get_player(self.protocol, name)
                        analyzer.hs, analyzer.bs, analyzer.ls = 0, 0, 0
            return connection.on_weapon_set(self, value)

        def on_disconnect(self):
            if self.name in list(self.protocol.analyzers.values()):
                for name in list(self.protocol.analyzers.keys()):
                    if self.protocol.analyzers[name] == self.name:
                        del self.protocol.analyzers[name]
                        analyzer = get_player(self.protocol, name)
                        if name != self.name:
                            analyzer.send_chat(
                                'You are no longer analyzing %s.  Player disconnected.' % (self.name))
            if self.name in self.protocol.analyzers:
                del self.protocol.analyzers[self.name]
            return connection.on_disconnect(self)

    class analyze_protocol(protocol):
        analyzers = {}

        def on_map_change(self, map):
            self.analyzers = {}
            protocol.on_map_change(self, map)
    return analyze_protocol, analyze_shotsConnection



"""
Detects and reacts to possible aimbot users.

Commands
^^^^^^^^

* ``/accuracy <player>`` shows player's accuracy per weapon
* ``/hackinfo <player>`` shows player's accuracy, K/D ratio and how often their cross-hair snaps onto another players head *admin only*

Options
^^^^^^^

.. code-block:: guess

   [aimbot]
   collect_data = true # saves hits and shots of each weapon to a csv file

.. codeauthor:: ?
"""
import os
import csv
import re
from math import sqrt, cos, pi

from twisted.internet.task import LoopingCall
from twisted.internet import reactor

from pyspades.constants import (
    WEAPON_TOOL, WEAPON_KILL, HEADSHOT_KILL,
    RIFLE_WEAPON, SMG_WEAPON, SHOTGUN_WEAPON,
)
from piqueserver.commands import command, admin, get_player
from piqueserver.config import config
from piqueserver.utils import parse

DISABLED, KICK, BAN, WARN_ADMIN = range(4)

# This controls which detection methods are enabled. If a player is detected
# using one of these methods, the player is kicked.
HEADSHOT_SNAP = WARN_ADMIN
HIT_PERCENT = WARN_ADMIN
KILLS_IN_TIME = WARN_ADMIN
MULTIPLE_BULLETS = WARN_ADMIN

# Minimum amount of time that must pass between admin warnings that are
# triggered by the same detection method. Time is in seconds.
WARN_INTERVAL_MINIMUM = 300

# These controls are only used if banning is enabled
# Time is given in minutes. Set to 0 for a permaban
HEADSHOT_SNAP_BAN_DURATION = parse("23hours")
HIT_PERCENT_BAN_DURATION = parse("1day")
KILLS_IN_TIME_BAN_DURATION = parse("2day")
MULTIPLE_BULLETS_BAN_DURATION = parse("1week")

# If more than or equal to this number of weapon hit packets are received
# from the client in half the weapon delay time, then an aimbot is detected.
# This method of detection should have 100% detection and no false positives
# with the current aimbot.
# Note that the current aimbot does not modify the number of bullets
# of the shotgun, so this method will not work if the player uses a shotgun.
# These values may need to be changed if an update to the aimbot is released.
RIFLE_MULTIPLE_BULLETS_MAX = 8
SMG_MULTIPLE_BULLETS_MAX = 8

# The minimum number of near misses + hits that are fired before kicking,
# banning, or warning an admin about someone using the hit percentage check
RIFLE_KICK_MINIMUM = 45
SMG_KICK_MINIMUM = 90
SHOTGUN_KICK_MINIMUM = 45

# Kick, ban, or warn when the above minimum is met and the
# bullet hit percentage is greater than or equal to this amount
RIFLE_KICK_PERC = 0.90
SMG_KICK_PERC = 0.80
SHOTGUN_KICK_PERC = 0.90

# If a player gets more kills than the KILL_THRESHOLD in the given
# KILL_TIME, kick, ban, or warn. This check is performed every
# time somebody kills someone with a gun
KILL_TIME = 20.0
KILL_THRESHOLD = 15

# If the number of headshot snaps exceeds the HEADSHOT_SNAP_THRESHOLD in the
# given HEADSHOT_SNAP_TIME, kick, ban, or warn. This check is performed every
# time somebody performs a headshot snap
HEADSHOT_SNAP_TIME = 20.0
HEADSHOT_SNAP_THRESHOLD = 6

# When the user's orientation angle (degrees) changes more than this amount,
# check if the user snapped to an enemy's head. If it is aligned with a head,
# record this as a headshot snap
HEADSHOT_SNAP_ANGLE = 90.0

# A near miss occurs when the player is NEAR_MISS_ANGLE degrees or less off
# of an enemy
NEAR_MISS_ANGLE = 10.0

# Approximate size of player's heads in blocks
HEAD_RADIUS = 0.7

# 128 is the approximate fog distance, but bump it up a little
# just in case
FOG_DISTANCE = 135.0

# Don't touch any of this stuff
FOG_DISTANCE2 = FOG_DISTANCE**2
NEAR_MISS_COS = cos(NEAR_MISS_ANGLE * (pi / 180.0))
HEADSHOT_SNAP_ANGLE_COS = cos(HEADSHOT_SNAP_ANGLE * (pi / 180.0))

aimbot_pattern = re.compile(".*(aim|bot|ha(ck|x)|cheat).*", re.IGNORECASE)
aimbot_config = config.section("aimbot")
collect_data = aimbot_config.option("collect_data", False)
config_dir = config.config_dir
def point_distance2(c1, c2):
    if c1.world_object is not None and c2.world_object is not None:
        p1 = c1.world_object.position
        p2 = c2.world_object.position
        return (p1.x - p2.x)**2 + (p1.y - p2.y)**2 + (p1.z - p2.z)**2


def dot3d(v1, v2):
    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]


def magnitude(v):
    return sqrt(v[0]**2 + v[1]**2 + v[2]**2)


def scale(v, factor):
    return (v[0] * factor, v[1] * factor, v[2] * factor)


def subtract(v1, v2):
    return (v1[0] - v2[0], v1[1] - v2[1], v1[2] - v2[2])


@command()
def accuracy(connection, name=None):
    if name is None:
        player = connection
    else:
        player = get_player(connection.protocol, name)
    return accuracy_player(player)


def accuracy_player(player, name_info=True):
    if player.rifle_count != 0:
        rifle_percent = "{0:.1f}%".format(
            (player.rifle_hits / player.rifle_count) * 100)
    else:
        rifle_percent = 'None'
    if player.smg_count != 0:
        smg_percent = "{0:.1f}%".format(
            (player.smg_hits / player.smg_count) * 100)
    else:
        smg_percent = 'None'
    if player.shotgun_count != 0:
        shotgun_percent = "{0:.1f}%".format(
            (player.shotgun_hits / player.shotgun_count) * 100)
    else:
        shotgun_percent = 'None'

    s = ''
    if name_info:
        s += player.name + ' has an accuracy of: '
    s += 'Rifle: %s SMG: %s Shotgun: %s.' % (
        rifle_percent, smg_percent, shotgun_percent)
    return s


@command(admin_only=True)
def hackinfo(connection, name):
    player = get_player(connection.protocol, name)
    return hackinfo_player(player)


def hackinfo_player(player):
    info = "%s #%s (%s) has an accuracy of: " % (
        player.name, player.player_id, player.address[0])
    info += accuracy_player(player, False)
    ratio = player.ratio_kills / float(max(1, player.ratio_deaths))
    info += " Kill-death ratio of %.2f (%s kills, %s deaths)." % (
        ratio, player.ratio_kills, player.ratio_deaths)
    info += " %i kills in the last %i seconds." % (
        player.get_kill_count(), KILL_TIME)
    info += " %i headshot snaps in the last %i seconds." % (
        player.get_headshot_snap_count(), HEADSHOT_SNAP_TIME)
    return info


def apply_script(protocol, connection, config):
    class Aimbot2Protocol(protocol):

        def start_votekick(self, payload):
            if not aimbot_pattern.match(payload.reason) is None:
                payload.target.warn_admin('Hack related votekick.')
            return protocol.start_votekick(self, payload)

    class Aimbot2Connection(connection):

        def __init__(self, *arg, **kw):
            connection.__init__(self, *arg, **kw)
            self.rifle_hits = self.smg_hits = self.shotgun_hits = 0
            self.rifle_count = self.smg_count = self.shotgun_count = 0
            self.last_target = None
            self.first_orientation = True
            self.kill_times = []
            self.headshot_snap_times = []
            self.bullet_loop = LoopingCall(self.on_bullet_fire)
            self.shot_time = 0.0
            self.multiple_bullets_count = 0
            self.headshot_snap_warn_time = self.hit_percent_warn_time = 0.0
            self.kills_in_time_warn_time = self.multiple_bullets_warn_time = 0.0

        def warn_admin(self, prefix='Possible aimbot detected.'):
            prefix += ' '
            message = hackinfo_player(self)
            for player in list(self.protocol.players.values()):
                if player.admin:
                    player.send_chat(prefix + message)
            irc_relay = self.protocol.irc_relay
            if irc_relay:
                if irc_relay.factory.bot and irc_relay.factory.bot.colors:
                    prefix = '\x0304' + prefix + '\x0f'
                irc_relay.send(prefix + message)

        def on_spawn(self, pos):
            self.first_orientation = True
            return connection.on_spawn(self, pos)

        def bullet_loop_start(self, interval):
            if not self.bullet_loop.running:
                self.bullet_loop.start(interval)

        def bullet_loop_stop(self):
            if self.bullet_loop.running:
                self.bullet_loop.stop()

        def get_headshot_snap_count(self):
            pop_count = 0
            headshot_snap_count = 0
            current_time = reactor.seconds()
            for old_time in self.headshot_snap_times:
                if current_time - old_time <= HEADSHOT_SNAP_TIME:
                    headshot_snap_count += 1
                else:
                    pop_count += 1

            self.headshot_snap_times = self.headshot_snap_times[pop_count:]
            return headshot_snap_count

        def on_orientation_update(self, x, y, z):
            if self.first_orientation or self.world_object is None:
                # we have no orientation data yet
                self.first_orientation = False
                return connection.on_orientation_update(self, x, y, z)

            orient = self.world_object.orientation
            old_orient_v = (orient.x, orient.y, orient.z)
            new_orient_v = (x, y, z)
            theta = dot3d(old_orient_v, new_orient_v)

            if theta > HEADSHOT_SNAP_ANGLE_COS:
                # The angle didn't change enough
                return connection.on_orientation_update(self, x, y, z)

            self_pos = self.world_object.position
            for enemy in self.team.other.get_players():
                enemy_pos = enemy.world_object.position
                position_v = (enemy_pos.x - self_pos.x, enemy_pos.y -
                              self_pos.y, enemy_pos.z - self_pos.z)
                c = scale(new_orient_v, dot3d(new_orient_v, position_v))
                h = magnitude(subtract(position_v, c))

                if h > HEAD_RADIUS:
                    # Didn't snap near a head
                    continue

                current_time = reactor.seconds()
                self.headshot_snap_times.append(current_time)

                if self.get_headshot_snap_count() < HEADSHOT_SNAP_THRESHOLD:
                    # Not enough headshots yet
                    continue

                if HEADSHOT_SNAP == BAN:
                    self.ban('Aimbot detected - headshot snap',
                             HEADSHOT_SNAP_BAN_DURATION)
                    return
                elif HEADSHOT_SNAP == KICK:
                    self.kick(
                        'Aimbot detected - headshot snap')
                    return
                elif HEADSHOT_SNAP == WARN_ADMIN:
                    if (current_time -
                            self.headshot_snap_warn_time) > WARN_INTERVAL_MINIMUM:
                        self.headshot_snap_warn_time = current_time
                        self.warn_admin()

            return connection.on_orientation_update(self, x, y, z)

        def on_shoot_set(self, shoot):
            if self.tool != WEAPON_TOOL:
                return connection.on_shoot_set(self, shoot)

            if shoot and not self.bullet_loop.running:
                self.possible_targets = []
                for enemy in self.team.other.get_players():
                    if point_distance2(self, enemy) <= FOG_DISTANCE2:
                        self.possible_targets.append(enemy)
                self.bullet_loop_start(self.weapon_object.delay)
            elif not shoot:
                self.bullet_loop_stop()
            return connection.on_shoot_set(self, shoot)

        def get_kill_count(self):
            current_time = reactor.seconds()
            kill_count = 0
            pop_count = 0
            for old_time in self.kill_times:
                if current_time - old_time <= KILL_TIME:
                    kill_count += 1
                else:
                    pop_count += 1

            self.kill_times = self.kill_times[pop_count:]
            return kill_count

        def on_kill(self, by, kill_type, grenade):
            if by is None or by is self:
                return connection.on_kill(self, by, kill_type, grenade)

            if kill_type != WEAPON_KILL and kill_type != HEADSHOT_KILL:
                return connection.on_kill(self, by, kill_type, grenade)

            by.kill_times.append(reactor.seconds())
            if by.get_kill_count() >= KILL_THRESHOLD:
                if KILLS_IN_TIME == BAN:
                    by.ban('Aimbot detected - kills in time window',
                           KILLS_IN_TIME_BAN_DURATION)
                    return
                elif KILLS_IN_TIME == KICK:
                    by.kick('Aimbot detected - kills in time window')
                    return
                elif KILLS_IN_TIME == WARN_ADMIN:
                    current_time = reactor.seconds()
                    if ((current_time - by.kills_in_time_warn_time)
                            > WARN_INTERVAL_MINIMUM):
                        by.kills_in_time_warn_time = current_time
                        by.warn_admin()

            return connection.on_kill(self, by, kill_type, grenade)

        def multiple_bullets_eject(self):
            if MULTIPLE_BULLETS == BAN:
                self.ban('Aimbot detected - multiple bullets',
                         MULTIPLE_BULLETS_BAN_DURATION)
            elif MULTIPLE_BULLETS == KICK:
                self.kick('Aimbot detected - multiple bullets')
            elif MULTIPLE_BULLETS == WARN_ADMIN:
                current_time = reactor.seconds()
                if ((current_time - self.multiple_bullets_warn_time)
                        > WARN_INTERVAL_MINIMUM):
                    self.multiple_bullets_warn_time = current_time
                    self.warn_admin()

        def on_hit(self, hit_amount, hit_player, hit_type, grenade):
            if self.team is hit_player.team or hit_type not in (
                    WEAPON_KILL, HEADSHOT_KILL):
                return connection.on_hit(
                    self, hit_amount, hit_player, hit_type, grenade)

            current_time = reactor.seconds()
            shotgun_use = False
            if current_time - \
                    self.shot_time > (0.5 * hit_player.weapon_object.delay):
                shotgun_use = True
                self.multiple_bullets_count = 0
                self.shot_time = current_time
            if hit_type == HEADSHOT_KILL:
                self.multiple_bullets_count += 1
            if self.weapon == RIFLE_WEAPON:
                self.rifle_hits += 1
                if self.multiple_bullets_count >= RIFLE_MULTIPLE_BULLETS_MAX:
                    self.multiple_bullets_eject()
                    return False
            elif self.weapon == SMG_WEAPON:
                self.smg_hits += 1
                if self.multiple_bullets_count >= SMG_MULTIPLE_BULLETS_MAX:
                    self.multiple_bullets_eject()
                    return False
            elif self.weapon == SHOTGUN_WEAPON:
                self.shotgun_hits += 1

            return connection.on_hit(
                self, hit_amount, hit_player, hit_type, grenade)

        def hit_percent_eject(self, hit_accuracy):
            message = 'Aimbot detected - %i%% %s hit accuracy' %\
                      (100.0 * hit_accuracy, self.weapon_object.name)
            if HIT_PERCENT == BAN:
                self.ban(message, HIT_PERCENT_BAN_DURATION)
            elif HIT_PERCENT == KICK:
                self.kick(message)
            elif HIT_PERCENT == WARN_ADMIN:
                current_time = reactor.seconds()
                if (current_time -
                        self.hit_percent_warn_time) > WARN_INTERVAL_MINIMUM:
                    self.hit_percent_warn_time = current_time
                    self.warn_admin()

        def check_percent(self):
            if self.weapon == RIFLE_WEAPON:
                rifle_perc = float(self.rifle_hits) / float(self.rifle_count)
                if self.rifle_count >= RIFLE_KICK_MINIMUM:
                    if rifle_perc >= RIFLE_KICK_PERC:
                        self.hit_percent_eject(rifle_perc)
            elif self.weapon == SMG_WEAPON:
                smg_perc = float(self.smg_hits) / float(self.smg_count)
                if self.smg_count >= SMG_KICK_MINIMUM:
                    if smg_perc >= SMG_KICK_PERC:
                        self.hit_percent_eject(smg_perc)
            elif self.weapon == SHOTGUN_WEAPON:
                shotgun_perc = float(self.shotgun_hits) / \
                    float(self.shotgun_count)
                if self.shotgun_count >= SHOTGUN_KICK_MINIMUM:
                    if shotgun_perc >= SHOTGUN_KICK_PERC:
                        self.hit_percent_eject(shotgun_perc)

        def on_bullet_fire(self):
            # Remembering the past offers a performance boost, particularly
            # with the SMG
            if self.last_target is not None:
                if self.last_target.hp is not None:
                    if self.check_near_miss(self.last_target):
                        self.check_percent()
                        return
            for enemy in self.possible_targets:
                if enemy.hp is not None and enemy is not self.last_target:
                    if self.check_near_miss(enemy):
                        self.last_target = enemy
                        self.check_percent()
                        return

        def check_near_miss(self, target):
            if self.world_object is not None and target.world_object is not None:
                p_self = self.world_object.position
                p_targ = target.world_object.position
                position_v = (p_targ.x - p_self.x, p_targ.y -
                              p_self.y, p_targ.z - p_self.z)
                orient = self.world_object.orientation
                orient_v = (orient.x, orient.y, orient.z)
                position_v_mag = magnitude(position_v)
                if position_v_mag != 0 and (
                    dot3d(
                        orient_v,
                        position_v) /
                        position_v_mag) >= NEAR_MISS_COS:
                    if self.weapon == RIFLE_WEAPON:
                        self.rifle_count += 1
                    elif self.weapon == SMG_WEAPON:
                        self.smg_count += 1
                    elif self.weapon == SHOTGUN_WEAPON:
                        self.shotgun_count += 1
                    return True
            return False

        # Data collection stuff
        def on_disconnect(self):
            self.bullet_loop_stop()
            if collect_data.get():
                if self.name is not None:
                    with open(os.path.join(config_dir,'aimbot2log.csv'), 'a+') as csvfile:
                        csvfile.seek(0)
                        fieldnames = ['name', 'rifle_hits', 'rifle_count', 'smg_hits', 'smg_count', 'shotgun_hits', 'shotgun_count']
                        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
                        try:
                            has_header = csv.Sniffer().has_header(csvfile.readline())
                        except csv.Error:
                            # Empty file causes this error
                            has_header = False
                        if not has_header:
                            writer.writeheader()
                        writer.writerow({
                            'name': self.name,
                            'rifle_hits': self.rifle_hits,
                            'rifle_count': self.rifle_count,
                            'smg_hits': self.smg_hits,
                            'smg_count': self.smg_count,
                            'shotgun_hits': self.shotgun_hits,
                            'shotgun_count': self.shotgun_count
                        })
                        csvfile.close()
            return connection.on_disconnect(self)

    return Aimbot2Protocol, Aimbot2Connection




import os
import filecmp
import shutil
import sys
import argparse
import tarfile
import json

from piqueserver.config import (config, TOML_FORMAT, JSON_FORMAT,
                                MAXMIND_DOWNLOAD, MAXMIND_DOWNLOAD_MD5,
                                SUPPORTED_PYTHONS)
import urllib.request
import hashlib

PKG_NAME = 'piqueserver'

def get_git_rev():
    if not os.path.exists(".git"):
        return 'snapshot'

    from distutils.spawn import find_executable
    if find_executable("git") is None:
        return 'gitless'

    import subprocess
    pipe = subprocess.Popen(
        ["git", "rev-parse", "HEAD"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)
    ret = pipe.stdout.read()[:40]
    if not ret:
        return 'unknown'
    return ret


def copytree(src, dst):
    """
    A re-implementation of shutil.copytree that doesn't fail if dst already exists.
    Other properties:
    Doesn't over-write if src/dst files don't differ.
    Creates a backup of dst file before over-writing.
    """
    os.makedirs(dst, exist_ok=True)
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            copytree(s, d)
        else:
            # create/copy if not exist
            if not os.path.exists(d):
                shutil.copy2(s, d)
            # if src/dst files differ, backup and over-write dst file
            elif not filecmp.cmp(s, d):
                shutil.copy2(d, d + '.bak')
                shutil.copy2(s, d)
            # skip if unchanged
            else:
                pass


def copy_config():
    config_source = os.path.dirname(os.path.abspath(__file__)) + '/config'
    print('Attempting to copy example config to %s (origin: %s).' %
          (config.config_dir, config_source))
    try:
        copytree(config_source, config.config_dir)
    except Exception as e:  # pylint: disable=broad-except
        print(e)
        return 1

    print('Complete! Please edit the files in %s to your liking.' %
          config.config_dir)
    return 0


def update_geoip(target_dir):
    db_filename = 'GeoLite2-City.mmdb'
    working_directory = os.path.join(target_dir, 'data/')
    zipped_path = os.path.join(working_directory,
                               os.path.basename(MAXMIND_DOWNLOAD))
    extracted_path = os.path.join(working_directory, db_filename)

    if not os.path.exists(target_dir):
        print('Configuration directory does not exist')
        return 1

    os.makedirs(working_directory, exist_ok=True)

    print('Downloading %s' % MAXMIND_DOWNLOAD)
    file_data = urllib.request.urlopen(MAXMIND_DOWNLOAD).read()

    print('Downloading %s' % MAXMIND_DOWNLOAD_MD5)
    sum_data = urllib.request.urlopen(MAXMIND_DOWNLOAD_MD5).read()
    print('Download Complete')

    # Both files are downloaded, but not stored before integrity check
    print('Checking integrity...')
    downloaded_sum = sum_data.decode()
    calculated_sum = hashlib.md5(file_data).hexdigest()
    if calculated_sum != downloaded_sum:
        print('md5 sums do not match')
        return 1

    print('OK')
    print('Saving file...')
    with open(zipped_path, 'wb') as f:
        f.write(file_data)

    print('Unpacking...')
    with tarfile.open(zipped_path, 'r:gz') as tar:
        comp_path = os.path.join(tar.next().name, db_filename)
        db = tar.extractfile(comp_path)
        with open(extracted_path, 'wb') as ex:
            ex.write(db.read())

    print('Unpacking Complete')

    print('Cleaning up...')
    os.remove(zipped_path)

    return 0


def main():
    if (sys.version_info.major, sys.version_info.minor) not in SUPPORTED_PYTHONS:
        print('Warning: you are running on an unsupported Python version.\n'
              'The server may not run correctly.\n'
              'Please see https://piqueserver.readthedocs.io/en/v1.0.0/supported-python-environments.html for more information.')

    description = '%s is an open-source Python server implementation ' \
                  'for the voxel-based game "Ace of Spades".' % PKG_NAME
    arg_parser = argparse.ArgumentParser(
        prog=PKG_NAME, description=description)

    arg_parser.add_argument(
        '-c',
        '--config-file',
        default=None,
        help='specify the config file - '
        'default is "config.toml" in the config dir')

    arg_parser.add_argument(
        '-j',
        '--json-parameters',
        help='add extra settings in json format '
        '(overrides the config present in the config file)')

    arg_parser.add_argument(
        '-d',
        '--config-dir',
        default=config.config_dir,
        help='specify the directory which contains '
        'maps, scripts, etc (in correctly named '
        'subdirs) - default is %s' % config.config_dir)

    arg_parser.add_argument(
        '--copy-config',
        action='store_true',
        help='copies the default/example config dir to '
        'its default location or as specified by "-d"')

    arg_parser.add_argument(
        '--update-geoip',
        action='store_true',
        help='download the latest geoip database')

    arg_parser.add_argument(
        '--version',
        action='store_true',
        help='show the version and exit')

    args = arg_parser.parse_args()

    # update the config_dir from cli args
    config.config_dir = args.config_dir

    # run the required tasks if args given
    if args.copy_config or args.update_geoip:
        if args.copy_config:
            status = copy_config()
            if status != 0:
                sys.exit(status)

        if args.update_geoip:
            status = update_geoip(config.config_dir)
            if status != 0:
                sys.exit(status)

        return  # if we have done a task, don't run the server

    if args.version:
        import piqueserver
        print("piqueserver", piqueserver.__version__)
        return

    # TODO: set config/map/script/log/etc. dirs from config file, thus removing
    # the need for the --config-dir argument and the config file is then a
    # single source of configuration

    # find and load the config
    # search order:
    # - --config-file (must have toml or json file extension)
    # - --config-dir/config.toml
    # - --config-dir/config.json
    # - ~/.config/piqueserver/config.toml
    # - ~/.config/piqueserver/config.json
    format_ = None
    if args.config_file is None:
        for format__, ext in ((TOML_FORMAT, 'toml'), (JSON_FORMAT, 'json')):
            config_file = os.path.join(config.config_dir,
                                       'config.{}'.format(ext))
            format_ = format__
            if os.path.exists(config_file):
                break
    else:
        config_file = args.config_file
        ext = os.path.splitext(config_file)[1]
        if ext == '.json':
            format_ = JSON_FORMAT
        elif ext == '.toml':
            format_ = TOML_FORMAT
        else:
            raise ValueError(
                'Unsupported config file format! Must have json or toml extension.'
            )

    config.config_file = config_file
    print('Loading config from {!r}'.format(config_file))
    try:
        with open(config_file) as fobj:
            config.load_from_file(fobj, format_=format_)
    except FileNotFoundError as e:
        print("Could not open Config file")
        print(e)
        return e.errno

    # update config with cli overrides
    if args.json_parameters:
        config.update_from_dict(json.loads(args.json_parameters))
    # We need to install the asyncio reactor before 
    # we add any imports like `twisted.internet.*` which install the default reactor.
    # We keep it here and not at package level to avoid installing the reactor more than once.
    # Twisted throws an exception if you install the reactor more than once.
    import asyncio
    from twisted.internet import asyncioreactor
    asyncioreactor.install(asyncio.get_event_loop())

    from piqueserver import server
    server.run()


if __name__ == "__main__":
    main()



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

MASTER_VERSION = 31
GAME_VERSION = 3
RIFLE_WEAPON, SMG_WEAPON, SHOTGUN_WEAPON = range(3)
TORSO, HEAD, ARMS, LEGS, MELEE = range(5)
SPADE_TOOL, BLOCK_TOOL, WEAPON_TOOL, GRENADE_TOOL = range(4)
# BUILD_BLOCK: place a single block
# DESTROY_BLOCK: destroy a single block
# SPADE_DESTROY: destroy the block as well as the blocks directly above and
#   below it (spade right-click)
# GRENADE_DESTROY: destroy a 3x3 area around the block
BUILD_BLOCK, DESTROY_BLOCK, SPADE_DESTROY, GRENADE_DESTROY = range(4)
BLUE_FLAG, GREEN_FLAG, BLUE_BASE, GREEN_BASE = range(4)
CHAT_ALL, CHAT_TEAM, CHAT_SYSTEM = range(3)
(WEAPON_KILL, HEADSHOT_KILL, MELEE_KILL, GRENADE_KILL, FALL_KILL,
    TEAM_CHANGE_KILL, CLASS_CHANGE_KILL) = range(7)
(ERROR_UNDEFINED, ERROR_BANNED, ERROR_TOO_MANY_CONNECTIONS, ERROR_WRONG_VERSION,
    ERROR_FULL, ERROR_SHUTDOWN) = range(6)
ERROR_KICKED, ERROR_INVALID_NAME = 10, 20

CTF_MODE, TC_MODE = range(2)
TC_CAPTURE_DISTANCE = 16  # 16 blocks
TC_CAPTURE_RATE = 0.05
MIN_TERRITORY_COUNT = 3
MAX_TERRITORY_COUNT = 7
NEUTRAL_TEAM = 2
SPAWN_RADIUS = 32
MINE_RANGE = 3
BUILD_TOLERANCE = 5

MELEE_DISTANCE = 3

MAX_CHAT_SIZE = 90  # more like 95, but just to make sure

RUBBERBAND_DISTANCE = 3

MAX_TIMER_SPEED = 2000
TIMER_WINDOW_ENTRIES = 40
MAX_RAPID_SPEED = 60  # 1 minute
RAPID_WINDOW_ENTRIES = 10

POSITION_RATE = 1
MAX_POSITION_RATE = 0.7

UPDATE_FPS = 60.0
UPDATE_FREQUENCY = 1 / UPDATE_FPS
NETWORK_FPS = 10.0

MIN_BLOCK_INTERVAL = 0.1
MAX_BLOCK_DISTANCE = 6
MAX_DIG_DISTANCE = 6
HIT_TOLERANCE = 5.0
CLIP_TOLERANCE = 10

TOOL_INTERVAL = {
    SPADE_TOOL: 0.1,
    BLOCK_TOOL: 0.1,
    GRENADE_TOOL: 1.0
}

WEAPON_INTERVAL = {
    RIFLE_WEAPON: 0.2,
    SMG_WEAPON: 0.05,
    SHOTGUN_WEAPON: 0.3
}



"""
Killing KILL_REQUIREMENT other players in under TIME_REQUIREMENT seconds
sends you into a rampage lasting RAMPAGE_DURATION seconds.
The rampage refills and drastically increases your weapons' rate of fire.

By default this means 3 kills in under 8 seconds to activate. For reference,
lines disappear from the killfeed 10 seconds after they appear.

Intended for use in frantic last team standing or free for all matches.

.. codeauthor:: hompy
"""

from collections import deque
from twisted.internet.reactor import callLater, seconds
from twisted.internet.task import LoopingCall
from pyspades import contained as loaders
from pyspades.common import make_color
from pyspades.constants import GRENADE_KILL, RIFLE_WEAPON, SMG_WEAPON, SHOTGUN_WEAPON

KILL_REQUIREMENT = 3
TIME_REQUIREMENT = 8.0
GRENADE_KILLS_COUNT = True
RAMPAGE_REFILLS = True
RAMPAGE_RELOADS = True
RAMPAGE_DURATION = 20.0
RAPID_INTERVALS = {
    RIFLE_WEAPON: 0.16,
    SMG_WEAPON: 0.08,
    SHOTGUN_WEAPON: 0.18
}
RAMPAGE_FOG_COLOR = (255, 0, 0)


def RAMPAGE_FOG_FUNC(): return RAMPAGE_FOG_COLOR


ANNOUNCE_RAMPAGE = True
S_RAMPAGE_START = '{player} IS ON A RAMPAGE!!'
S_RAMPAGE_KILLED = "{victim}'s rampage was ended by {killer}"


def resend_tool(player):
    set_tool = loaders.SetTool()
    set_tool.player_id = player.player_id
    set_tool.value = player.tool
    if player.weapon_object.shoot:
        player.protocol.broadcast_contained(set_tool)
    else:
        player.send_contained(set_tool)


def rapid_cycle(player):
    resend_tool(player)
    if not player.weapon_object.shoot:
        player.rampage_rapid_loop.stop()


def send_fog(player, color):
    fog_color = loaders.FogColor()
    fog_color.color = make_color(*color)
    player.send_contained(fog_color)


def fog_switch(player, colorgetter_a, colorgetter_b):
    if player.rampage:
        send_fog(player, colorgetter_a())
        player.rampage_warning_call = callLater(0.5, fog_switch, player,
                                                colorgetter_b, colorgetter_a)


def apply_script(protocol, connection, config):
    class RampageConnection(connection):
        rampage = False
        rampage_rapid_loop = None
        rampage_call = None
        rampage_warning_call = None
        rampage_kills = None
        rampage_reenable_rapid_hack_detect = None

        def start_rampage(self):
            self.rampage = True
            self.rampage_kills.clear()
            self.rampage_reenable_rapid_hack_detect = self.rapid_hack_detect
            self.rapid_hack_detect = False
            self.rampage_call = callLater(RAMPAGE_DURATION, self.end_rampage)
            if RAMPAGE_DURATION > 4.0:
                self.rampage_warning_call = callLater(
                    RAMPAGE_DURATION - 3.0,
                    fog_switch,
                    self,
                    self.protocol.get_fog_color,
                    RAMPAGE_FOG_FUNC)
            if RAMPAGE_REFILLS:
                self.refill()
            if RAMPAGE_RELOADS:
                weapon = self.weapon_object
                was_shooting = weapon.shoot
                weapon.reset()
                weapon_reload = loaders.WeaponReload()
                weapon_reload.player_id = self.player_id
                weapon_reload.clip_ammo = weapon.current_ammo
                weapon_reload.reserve_ammo = weapon.current_stock
                weapon.set_shoot(was_shooting)
                self.send_contained(weapon_reload)
            send_fog(self, RAMPAGE_FOG_COLOR)
            if ANNOUNCE_RAMPAGE:
                message = S_RAMPAGE_START.format(player=self.name)
                self.protocol.send_chat(message, global_message=None)

        def end_rampage(self):
            self.rampage = False
            self.rapid_hack_detect = self.rampage_reenable_rapid_hack_detect
            if self.rampage_call and self.rampage_call.active():
                self.rampage_call.cancel()
            self.rampage_call = None
            if self.rampage_warning_call and self.rampage_warning_call.active():
                self.rampage_warning_call.cancel()
            self.rampage_warning_call = None
            if self.rampage_rapid_loop and self.rampage_rapid_loop.running:
                self.rampage_rapid_loop.stop()
            send_fog(self, self.protocol.fog_color)

        def on_connect(self):
            self.rampage_rapid_loop = LoopingCall(rapid_cycle, self)
            self.rampage_kills = deque(maxlen=KILL_REQUIREMENT)
            connection.on_connect(self)

        def on_disconnect(self):
            if self.rampage:
                self.end_rampage()
            self.rampage_rapid_loop = None
            connection.on_disconnect(self)

        def on_reset(self):
            if self.rampage:
                self.end_rampage()
            connection.on_reset(self)

        def on_kill(self, killer, type, grenade):
            was_rampaging = self.rampage
            if self.rampage:
                self.end_rampage()
            if killer is not None and killer is not self:
                if was_rampaging and ANNOUNCE_RAMPAGE:
                    message = S_RAMPAGE_KILLED.format(victim=self.name,
                                                      killer=killer.name)
                    self.protocol.send_chat(message, global_message=None)
                if (not killer.rampage and killer.hp and
                    killer.team is not self.team and
                        (GRENADE_KILLS_COUNT or type != GRENADE_KILL)):
                    now = seconds()
                    killer.rampage_kills.append(now)
                    if (len(killer.rampage_kills) == KILL_REQUIREMENT and
                            killer.rampage_kills[0] >= now - TIME_REQUIREMENT):
                        killer.start_rampage()
            return connection.on_kill(self, killer, type, grenade)

        def on_grenade_thrown(self, grenade):
            if self.rampage:
                resend_tool(self)
            connection.on_grenade_thrown(self, grenade)

        def on_shoot_set(self, fire):
            if (self.rampage and fire and
                    self.rampage_rapid_loop and not self.rampage_rapid_loop.running):
                interval = RAPID_INTERVALS[self.weapon]
                self.rampage_rapid_loop.start(interval, now=False)
            connection.on_shoot_set(self, fire)

    def send_fog_rule(player): return not player.rampage

    class RampageProtocol(protocol):

        def set_fog_color(self, color):
            self.fog_color = color
            fog_color = loaders.FogColor()
            fog_color.color = make_color(*color)
            self.send_contained(fog_color, save=True, rule=send_fog_rule)

    return RampageProtocol, RampageConnection



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

"""
The ByteReader/Bytewriter classes are used to read and write various data types
from and to byte-like objects. This is used e.g. to read the contents of
packets.
"""

cdef extern from "bytes_c.cpp":
    char read_byte(char * data)
    unsigned char read_ubyte(char * data)
    short read_short(char * data, int big_endian)
    unsigned short read_ushort(char * data, int big_endian)
    int read_int(char * data, int big_endian)
    unsigned int read_uint(char * data, int big_endian)
    double read_float(char * data, int big_endian)
    char * read_string(char * data)

    stringstream * create_stream()
    void delete_stream(stringstream * stream)
    void write_byte(stringstream * stream, char value)
    void write_ubyte(stringstream * stream, unsigned char value)
    void write_short(stringstream * stream, short value, int big_endian)
    void write_ushort(stringstream * stream, unsigned short value, int big_endian)
    void write_int(stringstream * stream, int value, int big_endian)
    void write_uint(stringstream * stream, unsigned int value, int big_endian)
    void write_float(stringstream * stream, double value, int big_endian)
    void write_string(stringstream * stream, char * data, size_t size)
    void write(stringstream * stream, char * data, size_t size)
    void rewind_stream(stringstream * stream, int bytecount)
    object get_stream(stringstream * stream)
    size_t get_stream_size(stringstream * stream)
    size_t get_stream_pos(stringstream * stream)

cdef extern from "<sstream>" namespace "std":
    cdef cppclass stringstream:
        pass

class NoDataLeft(Exception):
    pass

DEF INT_ERROR = -0xFFFFFFFF >> 1
DEF LONG_LONG_ERROR = -0xFFFFFFFFFFFFFFFF >> 1
DEF FLOAT_ERROR = float('nan')

cdef class ByteReader:
    """Reads various data types from a bytes-like object"""
    def __init__(self, input_data, int start = 0, int size = -1):
        self.input = input_data
        self.data = input_data
        self.data += start
        self.pos = self.data
        if size == -1:
            size = len(input_data) - start
        self.size = size
        self.end = self.data + size
        self.start = start

    cdef char * check_available(self, int size) except NULL:
        cdef char * data = self.pos
        if data + size > self.end:
            raise NoDataLeft('not enough data')
        self.pos += size
        return data

    cpdef read(self, int bytecount = -1):
        """read a number of bytes

        Arguments:
            bytecount (int, optional): The number of bytes to read. If omitted, all bytes available are read

        Returns:
            bytes: ``bytecount`` bytes of data
        """
        cdef int left = self.dataLeft()
        if bytecount == -1 or bytecount > left:
            bytecount = left
        ret = self.pos[:bytecount]
        self.pos += bytecount
        return ret

    cpdef int readByte(self, bint unsigned = False) except INT_ERROR:
        """read one byte of data as integer

        Arguments:
            unsigned (bool): If true, interpret the byte as unsigned

        Returns:
            int: The value of the byte as int
        """
        cdef char * pos = self.check_available(1)
        if unsigned:
            return read_ubyte(pos)
        else:
            return read_byte(pos)

    cpdef int readShort(self, bint unsigned = False, bint big_endian = True) \
                        except INT_ERROR:
        """read two bytes of data as integer

        Arguments:
            unsigned (bool): If true, interpret the bytes as unsigned
            big_endian (bool, optional): If true, interpret the bytes as big endian

        Returns:
            int: The value of the bytes as int
        """
        cdef char * pos = self.check_available(2)
        if unsigned:
            return read_ushort(pos, big_endian)
        else:
            return read_short(pos, big_endian)

    cpdef long long readInt(self, bint unsigned = False,
                            bint big_endian = True) except LONG_LONG_ERROR:
        """read four bytes of data as integer

        Arguments:
            unsigned (bool): If true, interpret the bytes as unsigned
            big_endian (bool, optional): If true, interpret the bytes as big endian

        Returns:
            int: The value of the bytes as int
        """
        cdef char * pos = self.check_available(4)
        if unsigned:
            return read_uint(pos, big_endian)
        else:
            return read_int(pos, big_endian)

    cpdef float readFloat(self, bint big_endian = True) except? FLOAT_ERROR:
        """read four bytes of data as floating point number

        Arguments:
            big_endian (bool, optional): If true, interpret the bytes as big endian

        Returns:
            float: The value of the bytes as float
        """
        cdef char * pos = self.check_available(4)
        return read_float(pos, big_endian)

    cpdef bytes readString(self, int size = -1):
        """read a string

        Arguments:
            size (int): If set, read ``size`` bytes, else read all bytes available

        Returns:
            bytes: The value of the bytes
        """
        value = self.pos
        if size == -1:
            size = len(value) + 1
        if size > self.end - self.pos:
            size = self.end - self.pos
            value = value[:size]
        self.pos += size
        return bytes(value)

    cpdef ByteReader readReader(self, int size = -1):
        cdef int left = self.dataLeft()
        if size == -1 or size > left:
            size = left
        cdef ByteReader reader = ByteReader(self.input,
            (self.pos - self.data) + self.start, size)
        self.pos += size
        return reader

    cpdef size_t tell(self):
        """get the current position in the buffer

        Returns:
            int: The current position in bytes"""
        return self.pos - self.data

    cpdef int dataLeft(self):
        """get the number of bytes left in the buffer

        Returns:
            int: The number of bytes left"""
        return self.end - self.pos

    cpdef seek(self, size_t pos):
        """move to a position in the buffer

        Arguments:
            pos (int): position to seek to
        """
        self.pos = self.data + pos
        if self.pos > self.end:
            self.pos = self.end
        if self.pos < self.data:
            self.pos = self.data

    cdef void _skip(self, int bytecount):
        self.pos += bytecount
        if self.pos > self.end:
            self.pos = self.end
        if self.pos < self.data:
            self.pos = self.data

    cpdef skipBytes(self, int bytecount):
        """move the position ``bytecount`` bytes ahead

        Arguments:
            bytecount: number of bytes to move ahead
        """
        self._skip(bytecount)

    cpdef rewind(self, int value):
        """move the position ``bytecount`` bytes back

        Arguments:
            bytecount: number of bytes to move back
        """
        self._skip(-value)

    def __len__(self):
        return self.size

    def __bytes__(self):
        return self.data[:self.size]

cdef class ByteWriter:
    def __init__(self):
        self.stream = create_stream()

    cdef void writeSize(self, char * data, int size):
        write(self.stream, data, size)

    cpdef write(self, data):
        write(self.stream, data, len(data))

    cpdef writeByte(self, int value, bint unsigned = False):
        if unsigned:
            write_ubyte(self.stream, value)
        else:
            write_byte(self.stream, value)

    cpdef writeShort(self, int value, bint unsigned = False,
                     bint big_endian = True):
        if unsigned:
            write_ushort(self.stream, value, big_endian)
        else:
            write_short(self.stream, value, big_endian)

    cpdef writeInt(self, long long value, bint unsigned = False,
                   bint big_endian = True):
        if unsigned:
            write_uint(self.stream, value, big_endian)
        else:
            write_int(self.stream, value, big_endian)

    cpdef writeFloat(self, float value, bint big_endian = True):
        write_float(self.stream, value, big_endian)

    cpdef writeStringSize(self, char * value, int size):
        write_string(self.stream, value, size)

    cpdef writeString(self, value, int size = -1):
        write_string(self.stream, value, len(value))
        if size != -1:
            self.pad(size - (len(value) + 1))

    cpdef pad(self, int bytecount):
        cdef int i
        for i in range(bytecount):
            write_ubyte(self.stream, 0)

    cpdef rewind(self, int bytecount):
        rewind_stream(self.stream, bytecount)

    cpdef size_t tell(self):
        return get_stream_pos(self.stream)

    def __bytes__(self):
        return get_stream(self.stream)

    def __dealloc__(self):
        delete_stream(self.stream)

    def __len__(self):
        return get_stream_size(self.stream)



import random
import sys

from twisted.logger import Logger

from piqueserver.commands import command, join_arguments

log = Logger()


@command('servername', admin_only=True)
def server_name(connection, *arg):
    """
    Change the server name until it restarts
    /servername <new-name>
    Also affects the master list
    """
    if not arg:
        raise ValueError("no argument given")
    name = join_arguments(arg)
    protocol = connection.protocol
    protocol.set_server_name(name)
    message = "%s changed servername to '%s'" % (connection.name, name)
    log.info(message)
    connection.protocol.irc_say("* " + message)
    if connection in connection.protocol.players.values():
        return message
    return None


@command('server')
def server_info(connection):
    """
    Tell you the name of this server and its aos:// URI
    /server
    """
    protocol = connection.protocol
    msg = 'You are playing on "%s"' % protocol.name
    if protocol.identifier is not None:
        msg += ' at %s' % protocol.identifier
    return msg


@command()
def version(connection):
    """
    Tell you this server's piqueserver version
    /version
    """
    import piqueserver
    return 'Server version is "{}" on "{}"'.format(
        piqueserver.__version__, sys.platform)


@command()
def scripts(connection):
    """
    Tell you which scripts are enabled on this server currently
    /scripts
    """
    scripts = connection.protocol.config.get('scripts', [])
    if not scripts:
        return 'No scripts are enabled currently'
    return 'Scripts enabled: %s' % (', '.join(scripts))


@command('togglemaster', 'master', admin_only=True)
def toggle_master(connection):
    """
    Toggle connection to the master server list
    /togglemaster
    """
    protocol = connection.protocol
    protocol.set_master_state(not protocol.master)
    message = ("toggled master broadcast %s" % ['OFF', 'ON'][
        int(protocol.master)])
    protocol.irc_say("* %s " % connection.name + message)
    if connection in connection.protocol.players.values():
        return "You " + message



# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

from pyspades.bytes cimport ByteWriter, ByteReader

cdef class Loader:
    cpdef read(self, ByteReader reader)
    cpdef write(self, ByteWriter writer)
    cpdef ByteWriter generate(self)


# Copyright (c) Mathias Kaerlev 2011-2012.

# This file is part of pyspades.

# pyspades is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pyspades is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with pyspades.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=too-many-lines
"""
pyspades - default/featured server
"""
import asyncio
import itertools
import json
import os
import random
import sys
import time
from collections import deque
from ipaddress import AddressValueError, IPv4Address, ip_address, ip_network
from pprint import pprint
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple

import aiohttp
from enet import Address, Packet, Peer
from twisted.internet import reactor, threads
from twisted.internet.defer import Deferred, ensureDeferred, inlineCallbacks
from twisted.internet.task import LoopingCall, coiterate, deferLater
from twisted.internet.tcp import Port
from twisted.logger import (FilteringLogObserver, Logger, LogLevel,
                            LogLevelFilterPredicate, globalLogBeginner,
                            textFileLogObserver)
from twisted.python.logfile import DailyLogFile

# won't be used; just need to be executed
import piqueserver.core_commands  # pylint: disable=unused-import
from piqueserver import commands, extensions
from piqueserver.config import cast_duration, config
from piqueserver.console import create_console
from piqueserver.map import Map, MapNotFound, RotationInfo, check_rotation
from piqueserver.networkdict import NetworkDict
from piqueserver.player import FeatureConnection
from piqueserver.release import check_for_releases, format_release
from piqueserver.scheduler import Scheduler
from piqueserver.utils import as_deferred
from piqueserver.bansubscribe import bans_config_urls
from pyspades.bytes import NoDataLeft
from pyspades.constants import CTF_MODE, ERROR_SHUTDOWN, TC_MODE
from pyspades.master import MAX_SERVER_NAME_SIZE
from pyspades.server import ServerProtocol, Team
from pyspades.tools import make_server_identifier
from pyspades.vxl import VXLData

log = Logger()


def validate_team_name(name):
    if len(name) > 9:
        log.warn(
            "Team name's length exceeds 9 character limit. More info: https://git.io/fN2cI")
        # TODO: Once issue #345 is sorted out, we can do a proper validation
        # for now we just warn
        # return False
    return True

# TODO: move to a better place if reusable


def sleep(secs):
    return deferLater(reactor, secs, lambda: None)


# declare configuration options
bans_config = config.section('bans')
logging_config = config.section('logging')
team1_config = config.section('team1')
team2_config = config.section('team2')

bans_file = bans_config.option('file', default='bans.txt')
bans_urls = bans_config.option('urls', [])
respawn_time_option = config.option(
    'respawn_time', default="8sec", cast=cast_duration)
respawn_waves = config.option('respawn_waves', default=False)
game_mode = config.option('game_mode', default='ctf')
random_rotation = config.option('random_rotation', default=False)
passwords = config.option('passwords', default={})
logfile = logging_config.option('logfile', default='./logs/log.txt')
loglevel = logging_config.option('loglevel', default='info')
map_rotation = config.option('rotation', default=['classicgen', 'random'],
                             validate=lambda x: isinstance(x, list))
default_time_limit = config.option(
    'default_time_limit', default="20min",
    cast=lambda x: cast_duration(x)/60)
cap_limit = config.option('cap_limit', default=10,
                          validate=lambda x: isinstance(x, (int, float)))
advance_on_win = config.option('advance_on_win', default=False,
                               validate=lambda x: isinstance(x, bool))
everyone_is_admin = config.option('everyone_is_admin', default=False,
                               validate=lambda x: isinstance(x, bool))
team1_name = team1_config.option(
    'name', default='Blue', validate=validate_team_name)
team2_name = team2_config.option(
    'name', default='Green', validate=validate_team_name)
team1_color = team1_config.option('color', default=(0, 0, 196))
team2_color = team2_config.option('color', default=(0, 196, 0))
friendly_fire = config.option('friendly_fire', default=False)
friendly_fire_on_grief = config.option('friendly_fire_on_grief',
                                       default=True)
grief_friendly_fire_time = config.option('grief_friendly_fire_time',
                                         default='2sec', cast=cast_duration)
spade_teamkills_on_grief = config.option('spade_teamkills_on_grief',
                                         default=False)
time_announcements = config.option('time_announcements', default=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
                                                                  30, 60, 120, 180, 240, 300, 600,
                                                                  900, 1200, 1800, 2400, 3000])
rights = config.option('rights', default={})
port_option = config.option('port', default=32887,
                            validate=lambda n: isinstance(n, int))
fall_damage = config.option('fall_damage', default=True)
teamswitch_interval = config.option(
    'teamswitch_interval', default="0sec", cast=cast_duration)
teamswitch_allowed = config.option('teamswitch_allowed', default=True)
max_players = config.option('max_players', default=20)
melee_damage = config.option('melee_damage', default=100)
max_connections_per_ip = config.option('max_connections_per_ip', default=0)
server_prefix = config.option('server_prefix', default='[*]')
balanced_teams = config.option('balanced_teams', default=2)
login_retries = config.option('login_retries', 1)
default_ban_duration = bans_config.option(
    'default_duration', default="1day", cast=cast_duration)
speedhack_detect = config.option('speedhack_detect', True)
rubberband_distance = config.option('rubberband_distance', default=10)
user_blocks_only = config.option('user_blocks_only', False)
debug_log_enabled = logging_config.option('debug_log', False)
logging_profile_option = logging_config.option('profile', False)
set_god_build = config.option('set_god_build', False)
ssh_enabled = config.section('ssh').option('enabled', False)
irc_options = config.option('irc', {})
status_server_enabled = config.section(
    'status_server').option('enabled', False)
ban_publish = bans_config.option('publish', False)
ban_publish_port = bans_config.option('publish_port', 32885)
logging_rotate_daily = logging_config.option('rotate_daily', False)
tip_frequency = config.option(
    'tips_frequency', default="5sec", cast=lambda x: cast_duration(x)/60)
register_master_option = config.option('master', False)

default_ip_getter = 'https://services.buildandshoot.com/getip'
ip_getter_option = config.option('ip_getter', default_ip_getter)
name_option = config.option(
    'name', default='piqueserver #%s' % random.randrange(0, 2000))
motd_option = config.option('motd')
help_option = config.option('help', default=[
    'Server name: %(server_name)s',
    'Map: %(map_name)s by %(map_author)s',
    'Game mode: %(game_mode)s',
    '/commands Prints all available commands',
    '/help <command_name> Gives description and usage info for a command',
    '/help Prints this message',
])
rules_option = config.option('rules')
tips_option = config.option('tips')
network_interface = config.option('network_interface', default='')
scripts_option = config.option(
    'scripts', default=[], validate=extensions.check_scripts)


def ensure_dir_exists(filename: str) -> None:
    d = os.path.dirname(filename)
    os.makedirs(d, exist_ok=True)


def random_choice_cycle(choices):
    while True:
        yield random.choice(choices)


class FeatureTeam(Team):
    locked = False

    def get_entity_location(self, entity_id: int) -> Tuple[int, int, int]:
        get_location = self.protocol.map_info.get_entity_location
        if get_location is not None:
            result = get_location(self, entity_id)
            if result is not None:
                return result
        return Team.get_entity_location(self, entity_id)


class EndCall:
    _active = True

    def __init__(self, protocol, delay: int, func: Callable, *arg, **kw) -> None:
        self.protocol = protocol
        protocol.end_calls.append(self)
        self.delay = delay
        self.func = func
        self.arg = arg
        self.kw = kw
        self.call = None  # type: Deferred

    def set(self, value: Optional[float]) -> None:
        if value is None:
            if self.call is not None:
                self.call.cancel()
                self.call = None
        elif value is not None:
            value = value - self.delay
            if value <= 0.0:
                self.cancel()
            elif self.call:
                # In Twisted==18.9.0, reset() is broken when using
                # AsyncIOReactor
                # self.call.reset(value)
                self.call.cancel()
                self.call = reactor.callLater(value, self.fire)
            else:
                self.call = reactor.callLater(value, self.fire)

    def fire(self):
        self.call = None
        self.cancel()
        self.func(*self.arg, **self.kw)

    def cancel(self) -> None:
        self.set(None)
        self.protocol.end_calls.remove(self)
        self._active = False

    def active(self) -> bool:
        return self._active and (self.call and self.call.active())


class FeatureProtocol(ServerProtocol):
    connection_class = FeatureConnection
    bans = None
    ban_publish = None
    ban_manager = None
    everyone_is_admin = False
    player_memory = None
    irc_relay = None
    balanced_teams = None
    timestamps = None
    building = True
    killing = True
    global_chat = True
    remote_console = None
    debug_log = None
    advance_call = None
    master_reconnect_call = None
    master = False
    ip = None
    identifier = None

    planned_map = None

    map_info = None
    spawns = None
    user_blocks = None
    god_blocks = None

    last_time = None
    interface = None

    team_class = FeatureTeam

    game_mode = None  # default to None so we can check
    time_announce_schedule = None

    default_fog = (128, 232, 255)

    def __init__(self, interface: bytes, config_dict: Dict[str, Any]) -> None:
        # logfile path relative to config dir if not abs path
        log_filename = logfile.get()
        if log_filename.strip():  # catches empty filename
            if not os.path.isabs(log_filename):
                log_filename = os.path.join(config.config_dir, log_filename)
            ensure_dir_exists(log_filename)
            if logging_rotate_daily.get():
                logging_file = DailyLogFile(log_filename, '.')
            else:
                logging_file = open(log_filename, 'a')
            predicate = LogLevelFilterPredicate(
                LogLevel.levelWithName(loglevel.get()))
            observers = [
                FilteringLogObserver(
                    textFileLogObserver(sys.stderr), [predicate]),
                FilteringLogObserver(
                    textFileLogObserver(logging_file), [predicate])
            ]
            globalLogBeginner.beginLoggingTo(observers)
            log.info('piqueserver started on %s' % time.strftime('%c'))

        self.config = config_dict
        if random_rotation.get():
            self.map_rotator_type = random_choice_cycle
        else:
            self.map_rotator_type = itertools.cycle
        self.default_time_limit = default_time_limit.get()
        self.default_cap_limit = cap_limit.get()
        self.advance_on_win = int(advance_on_win.get())
        self.win_count = itertools.count(1)
        self.bans = NetworkDict()

        # attempt to load a saved bans list
        try:
            with open(os.path.join(config.config_dir, bans_file.get()), 'r') as f:
                self.bans.read_list(json.load(f))
            log.debug("loaded {count} bans", count=len(self.bans))
        except FileNotFoundError:
            log.debug("skip loading bans: file unavailable",
                      count=len(self.bans))
        except IOError as e:
            log.error('Could not read bans file ({}): {}'.format(bans_file.get(), e))
        except ValueError as e:
            log.error('Could not parse bans file ({}): {}'.format(bans_file.get(), e))

        self.hard_bans = set()  # possible DDoS'ers are added here
        self.player_memory = deque(maxlen=100)
        if len(self.name) > MAX_SERVER_NAME_SIZE:
            log.warn('(server name too long; it will be truncated to "%s")' % (
                self.name[:MAX_SERVER_NAME_SIZE]))
        self.respawn_time = respawn_time_option.get()
        self.respawn_waves = respawn_waves.get()

        # since AoS only supports CTF and TC at a protocol level, we need to get
        # the base game mode if we are using a custom game mode.
        game_mode_name = game_mode.get()
        if game_mode_name == 'ctf':
            self.game_mode = CTF_MODE
        elif game_mode.get() == 'tc':
            self.game_mode = TC_MODE
        elif self.game_mode not in [CTF_MODE, TC_MODE]:
            raise ValueError(
                'invalid game mode: custom game mode "{}" does not set '
                'protocol.game_mode to one of TC_MODE or CTF_MODE. Are '
                'you sure the thing you have specified is a game mode?'.format(
                    game_mode_name))

        self.game_mode_name = game_mode.get().split('.')[-1]
        self.team1_name = team1_name.get()[:9]
        self.team2_name = team2_name.get()[:9]
        self.team1_color = tuple(team1_color.get())
        self.team2_color = tuple(team2_color.get())
        self.friendly_fire = friendly_fire.get()
        self.friendly_fire_on_grief = friendly_fire_on_grief.get()
        self.friendly_fire_time = grief_friendly_fire_time.get()
        self.spade_teamkills_on_grief = spade_teamkills_on_grief.get()
        self.fall_damage = fall_damage.get()
        self.teamswitch_interval = teamswitch_interval.get()
        self.teamswitch_allowed = teamswitch_allowed.get()
        self.max_players = max_players.get()
        self.melee_damage = melee_damage.get()
        self.max_connections_per_ip = max_connections_per_ip.get()
        self.passwords = passwords.get()
        self.server_prefix = server_prefix.get()
        self.time_announcements = time_announcements.get()
        self.balanced_teams = balanced_teams.get()
        self.login_retries = login_retries.get()

        # voting configuration
        self.default_ban_time = default_ban_duration.get()

        self.speedhack_detect = speedhack_detect.get()
        self.rubberband_distance = rubberband_distance.get()
        if user_blocks_only.get():
            self.user_blocks = set()
        self.set_god_build = set_god_build.get()
        self.debug_log = debug_log_enabled.get()
        if self.debug_log:
            # TODO: make this configurable
            pyspades.debug.open_debug_log(
                os.path.join(config.config_dir, 'debug.log'))
        if ssh_enabled.get():
            from piqueserver.ssh import RemoteConsole
            self.remote_console = RemoteConsole(self)
        irc = irc_options.get()
        if irc.get('enabled', False):
            from piqueserver.irc import IRCRelay
            self.irc_relay = IRCRelay(self, irc)
        if status_server_enabled.get():
            from piqueserver.statusserver import StatusServer
            self.status_server = StatusServer(self)
            ensureDeferred(self.status_server.listen())
        if ban_publish.get():
            from piqueserver.banpublish import PublishServer
            self.ban_publish = PublishServer(self, ban_publish_port.get())
        if bans_config_urls.get():
            from piqueserver import bansubscribe
            self.ban_manager = bansubscribe.BanManager(self)
        self.start_time = time.time()
        self.end_calls = []
        # TODO: why is this here?
        create_console(self)

        for user_type, func_names in rights.get().items():
            for func_name in func_names:
                commands.add_rights(user_type, func_name)

        if everyone_is_admin.get():
            self.everyone_is_admin = True

        self.port = port_option.get()
        ServerProtocol.__init__(self, self.port, interface)
        self.host.intercept = self.receive_callback

        try:
            self.set_map_rotation(self.config['rotation'])
        except MapNotFound as e:
            log.critical('Invalid map in map rotation (%s), exiting.' % e.map)
            raise SystemExit

        map_load_d = self.advance_rotation()
        # discard the result of the map advance for now
        map_load_d.addCallback(lambda x: self._post_init())

        ip_getter = ip_getter_option.get()
        if ip_getter:
            ensureDeferred(as_deferred(self.get_external_ip(ip_getter)))

        self.new_release = None
        notify_new_releases = config.option(
            "release_notifications", default=True)
        if notify_new_releases.get():
            ensureDeferred(as_deferred(self.watch_for_releases()))

        self.vacuum_loop = LoopingCall(self.vacuum_bans)
        # Run the vacuum every 6 hours, and kick it off it right now
        self.vacuum_loop.start(60 * 60 * 6, True)

        reactor.addSystemEventTrigger(
            'before', 'shutdown', lambda: ensureDeferred(self.shutdown()))

    def _post_init(self):
        """called after the map has been loaded"""
        self.update_format()
        self.tip_frequency = tip_frequency.get()
        if self.tips is not None and self.tip_frequency > 0:
            reactor.callLater(self.tip_frequency * 60, self.send_tip)

        self.master = register_master_option.get()
        self.set_master()

    async def get_external_ip(self, ip_getter: str) -> Iterator[Deferred]:
        log.info(
            'Retrieving external IP from {!r} to generate server identifier.'.format(ip_getter))
        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(ip_getter) as response:
                    ip = await response.text()
                    ip = IPv4Address(ip.strip())
        except AddressValueError as e:
            log.warn('External IP getter service returned invalid data.\n'
                     'Please check the "ip_getter" setting in your config.')
            return
        except Exception as e:  # pylint: disable=broad-except
            log.warn("Getting external IP failed: {reason}", reason=e)
            return

        self.ip = ip
        self.identifier = make_server_identifier(ip, self.port)
        log.info('Server public ip address: {}:{}'.format(ip, self.port))
        log.info('Public aos identifier: {}'.format(self.identifier))

    def set_time_limit(self, time_limit: Optional[int] = None, additive:
                       bool = False) -> Optional[int]:
        advance_call = self.advance_call
        add_time = 0.0
        if advance_call is not None:
            add_time = ((advance_call.getTime() - reactor.seconds()) / 60.0)
            advance_call.cancel()
            self.advance_call = None
        time_limit = time_limit or self.default_time_limit
        if not time_limit:
            for call in self.end_calls[:]:
                call.set(None)
            return None

        if additive:
            time_limit = min(time_limit + add_time, self.default_time_limit)

        seconds = time_limit * 60
        self.advance_call = reactor.callLater(seconds, self._time_up)

        for call in self.end_calls[:]:
            call.set(seconds)

        if self.time_announce_schedule is not None:
            self.time_announce_schedule.reset()
        self.time_announce_schedule = Scheduler(self)
        for seconds in self.time_announcements:
            self.time_announce_schedule.call_end(seconds,
                                                 self._next_time_announce)

        return time_limit

    def _next_time_announce(self):
        remaining = self.advance_call.getTime() - reactor.seconds()
        if remaining < 60.001:
            if remaining < 10.001:
                self.send_chat('%s...' % int(round(remaining)))
            else:
                self.send_chat('%s seconds remaining.' % int(round(remaining)))
        else:
            self.send_chat('%s minutes remaining.' %
                           int(round(remaining / 60)))

    def _time_up(self):
        self.advance_call = None
        self.advance_rotation('Time up!')

    def advance_rotation(self, message: Optional[str] = None) -> Deferred:
        """
        Advances to the next map in the rotation. If message is provided
        it will send it to the chat, waits for 10 seconds and then advances.

        Returns:
            Deferred that fires when the map has been loaded
        """
        self.set_time_limit(False)
        if self.planned_map is None:
            self.planned_map = next(self.map_rotator)
        planned_map = self.planned_map
        self.planned_map = None
        self.on_advance(planned_map)

        async def do_advance():
            if message is not None:
                log.info("advancing to map '{name}' ({reason}) in 10 seconds",
                         name=planned_map.full_name, reason=message)
                self.send_chat(
                    '{} Next map: {}.'.format(message, planned_map.full_name),
                    irc=True)
                await sleep(10)
            else:
                log.info("advancing to map '{name}'",
                         name=planned_map.full_name)

            await self.set_map_name(planned_map)

        return ensureDeferred(do_advance())

    def get_mode_name(self) -> str:
        return self.game_mode_name

    async def set_map_name(self, rot_info: RotationInfo) -> None:
        """
        Sets the map by its name.
        """
        map_info = await self.make_map(rot_info)
        if self.map_info:
            self.on_map_leave()
        self.map_info = map_info
        self.max_score = self.map_info.cap_limit or self.default_cap_limit
        self.set_map(self.map_info.data)
        self.set_time_limit(self.map_info.time_limit)
        self.update_format()

    def set_server_name(self, name: str) -> None:
        name_option.set(name)
        self.update_format()

    def make_map(self, rot_info: RotationInfo) -> Deferred:
        """
        Creates and returns a Map object from rotation info in a new thread

        Returns:
            Deferred that resolves to a `Map` object.
        """
        # we must do this in a new thread, since map generation might take so
        # long that clients time out.
        return threads.deferToThread(
            Map, rot_info, os.path.join(config.config_dir, 'maps'))

    def set_map_rotation(self, maps: List[str]) -> None:
        """
        Over-writes the current map rotation with provided one.
        `FeatureProtocol.advance_rotation` still needs to be called to actually
        change the map,
        """
        maps = check_rotation(maps, os.path.join(config.config_dir, 'maps'))
        self.maps = maps
        self.map_rotator = self.map_rotator_type(maps)

    def get_map_rotation(self):
        return [map_item.full_name for map_item in self.maps]

    def is_indestructable(self, x: int, y: int, z: int) -> bool:
        if self.user_blocks is not None:
            if (x, y, z) not in self.user_blocks:
                return True
        if self.god_blocks is not None:
            if (x, y, z) in self.god_blocks:  # pylint: disable=unsupported-membership-test
                return True
        map_is_indestructable = self.map_info.is_indestructable
        if map_is_indestructable is not None:
            if map_is_indestructable(self, x, y, z):
                return True
        return False

    def update_format(self) -> None:
        """
        Called when the map (or other variables) have been updated
        """
        self.name = self.format(name_option.get())
        self.motd = self.format_lines(motd_option.get())
        self.help = self.format_lines(help_option.get())
        self.tips = self.format_lines(tips_option.get())
        self.rules = self.format_lines(rules_option.get())
        if self.master_connection is not None:
            self.master_connection.send_server()

    def format(self, value: str, extra: Optional[Dict[str, str]] = None) -> str:
        map_info = self.map_info
        format_dict = {
            'map_name': map_info.name,
            'map_author': map_info.author,
            'map_description': map_info.description,
            'game_mode': self.get_mode_name(),
            'server_name': self.name,
        }

        if extra:
            format_dict.update(extra)
        # format with both old-style and new string formatting to stay
        # compatible with older configs
        return value.format(**format_dict) % format_dict

    def format_lines(self, value: List[str]) -> List[str]:
        if value is None:
            return
        return [self.format(line) for line in value]

    def got_master_connection(self, client):
        log.info('Master connection established.')
        ServerProtocol.got_master_connection(self, client)

    def master_disconnected(self, client=None):
        ServerProtocol.master_disconnected(self, client)
        if self.master and self.master_reconnect_call is None:
            if client:
                message = 'Master connection could not be established'
            else:
                message = 'Master connection lost'
            log.info('%s, reconnecting in 60 seconds...' % message)
            self.master_reconnect_call = reactor.callLater(
                60, self.reconnect_master)

    def reconnect_master(self):
        self.master_reconnect_call = None
        self.set_master()

    def set_master_state(self, value):
        if value == self.master:
            return
        self.master = value
        has_connection = self.master_connection is not None
        has_reconnect = self.master_reconnect_call is not None
        if value:
            if not has_connection and not has_reconnect:
                self.set_master()
        else:
            if has_reconnect:
                self.master_reconnect_call.cancel()
                self.master_reconnect_call = None
            if has_connection:
                self.master_connection.disconnect()

    async def shutdown(self):
        """
        Notifies players and disconnects them before a shutdown.
        """
        if not self.connections:
            # exit instantly if nobody is connected anyway
            return

        # send shutdown notification
        log.info("disconnecting players")
        self.broadcast_chat("Server shutting down in 3sec.")
        for i in range(3, 0, -1):
            self.broadcast_chat(str(i)+"...")
            await sleep(1)

        # disconnect all players
        for connection in list(self.connections.values()):
            connection.disconnect(ERROR_SHUTDOWN)

        # give the connections some time to terminate
        await sleep(0.2)

    def add_ban(self, ip, reason, duration, name=None):
        """
        Ban an ip with an optional reason and duration in seconds. If duration
        is None, ban is permanent.
        """
        network = ip_network(str(ip), strict=False)
        for connection in list(self.connections.values()):
            if ip_address(connection.address[0]) in network:
                name = connection.name
                connection.kick(silent=True)
        if duration:
            duration = reactor.seconds() + duration
        else:
            duration = None
        self.bans[ip] = (name or '(unknown)', reason, duration)
        self.save_bans()

    def remove_ban(self, ip):
        results = self.bans.remove(ip)
        log.info('Removing ban: {ip} {results}',
                 ip=ip, results=results)
        self.save_bans()

    async def watch_for_releases(self):
        """Starts a loop for `check_for_releases` and updates `self.new_release`."""
        while True:
            self.new_release = await check_for_releases()
            if self.new_release:
                log.info("#" * 60)
                log.info(format_release(self.new_release))
                log.info("#" * 60)
            await asyncio.sleep(86400)  # 24 hrs

    def vacuum_bans(self):
        """remove any bans that might have expired. This takes a while, so it is
        split up over the event loop"""

        def do_vacuum_bans():
            """do the actual clearing of bans"""

            bans_count = len(self.bans)
            log.info("starting ban vacuum with {count} bans",
                     count=bans_count)
            start_time = time.time()

            # create a copy of the items, so we don't have issues modifying
            # while iteraing
            for ban in list(self.bans.iteritems()):
                ban_exipry = ban[1][2]
                if ban_exipry is None:
                    # entry never expires
                    continue
                if ban[1][2] < start_time:
                    # expired
                    del self.bans[ban[0]]
                yield
            log.debug("ban vacuum took {time:.2f} seconds, removed {count} bans",
                      count=bans_count - len(self.bans),
                      time=time.time() - start_time)
            self.save_bans()

        # TODO: use cooperate() here instead, once you figure out why it's
        # swallowing errors. Perhaps try add an errback?
        coiterate(do_vacuum_bans())

    def undo_last_ban(self):
        result = self.bans.pop()
        self.save_bans()
        return result

    def save_bans(self):
        ban_file = os.path.join(config.config_dir, bans_file.get())
        ensure_dir_exists(ban_file)

        start_time = reactor.seconds()
        with open(ban_file, 'w') as f:
            json.dump(self.bans.make_list(), f, indent=2)
        log.debug("saving {count} bans took {time:.2f} seconds",
                  count=len(self.bans),
                  time=reactor.seconds() - start_time)

        if self.ban_publish is not None:
            self.ban_publish.update()

    def receive_callback(self, address: Address, data: bytes) -> int:
        """This hook receives the raw UDP data before it is processed by enet"""

        # exceptions get swallowed in the pyenet C stuff, so we catch anything
        # for now. This should ideally get fixed in pyenet instead.
        try:
            # reply to ASCII HELLO messages with HI so that clients can measure the
            # connection latency
            if data == b'HELLO':
                self.host.socket.send(address, b'HI')
                return 1
            # reply to ASCII HELLOLAN messages with server data for LAN discovery
            elif data == b'HELLOLAN':
                # we might receive a HELLOLAN before the map has been loaded
                # if so, return a dummy string instead
                if self.map_info:
                    map_name = self.map_info.short_name
                else:
                    map_name = "loading..."
                entry = {
                    "name": self.name,
                    "players_current": self.get_player_count(),
                    "players_max": self.max_players,
                    "map": map_name,
                    "game_mode": self.get_mode_name(),
                    "game_version": "0.75"
                }
                payload = json.dumps(entry).encode()
                self.host.socket.send(address, payload)
                return 1

            # This drop the connection of any ip in hard_bans
            if address.host in self.hard_bans:
                return 1
        except Exception:
            import traceback
            traceback.print_exc()

        return 0

    def data_received(self, peer: Peer, packet: Packet) -> None:
        ip = peer.address.host
        current_time = reactor.seconds()
        try:
            ServerProtocol.data_received(self, peer, packet)
        except (NoDataLeft, ValueError):
            import traceback
            traceback.print_exc()
            log.info(
                'IP %s was hardbanned for invalid data or possibly DDoS.' % ip)
            self.hard_bans.add(ip)
            return
        dt = reactor.seconds() - current_time
        if dt > 1.0:
            log.warn('processing {!r} from {} took {}'.format(
                packet.data, ip, dt))

    def irc_say(self, msg: str, me: bool = False) -> None:
        if self.irc_relay:
            if me:
                self.irc_relay.me(msg, do_filter=True)
            else:
                self.irc_relay.send(msg, do_filter=True)

    def send_tip(self):
        line = self.tips[random.randrange(len(self.tips))]
        self.send_chat(line)
        reactor.callLater(self.tip_frequency * 60, self.send_tip)

    # pylint: disable=arguments-differ
    def broadcast_chat(self, value, global_message=True, sender=None,
                       team=None, irc=False):
        """
        Send a chat message to many users
        """
        if irc:
            self.irc_say('* %s' % value)
        ServerProtocol.send_chat(self, value, global_message, sender, team)

    # backwards compatability
    send_chat = broadcast_chat

    # log high CPU usage

    def update_world(self):
        last_time = self.last_time
        current_time = reactor.seconds()
        if last_time is not None:
            dt = current_time - last_time
            if dt > 1.0:
                log.warn('high CPU usage detected - %s' % dt)
        self.last_time = current_time
        ServerProtocol.update_world(self)
        time_taken = reactor.seconds() - current_time
        if time_taken > 1.0:
            log.warn(
                'World update iteration took %s, objects: %s' %
                (time_taken, self.world.objects))

    # events

    def on_map_change(self, the_map: VXLData) -> None:
        self.set_fog_color(
            getattr(self.map_info.info, 'fog', self.default_fog)
        )

        map_on_map_change = self.map_info.on_map_change
        if map_on_map_change is not None:
            map_on_map_change(self, the_map)

    def on_map_leave(self):
        map_on_map_leave = self.map_info.on_map_leave
        if map_on_map_leave is not None:
            map_on_map_leave(self)

    def on_game_end(self):
        if self.advance_on_win <= 0:
            self.irc_say('Round ended!', me=True)
        elif next(self.win_count) % self.advance_on_win == 0:
            self.advance_rotation('Game finished!')

    def on_advance(self, map_name: str) -> None:
        pass

    def on_ban_attempt(self, connection, reason, duration):
        return True

    def on_ban(self, connection, reason, duration):
        pass

    # voting

    def cancel_vote(self, connection=None):
        return 'No vote in progress.'

    # useful twisted wrappers

    def listenTCP(self, *arg, **kw) -> Port:
        return reactor.listenTCP(
            *arg, interface=network_interface.get(), **kw)

    def connectTCP(self, *arg, **kw):
        return reactor.connectTCP(
            *arg,
            bindAddress=(
                network_interface.get(),
                0),
            **kw)

    # before-end calls

    def call_end(self, delay: int, func: Callable, *arg, **kw) -> EndCall:
        call = EndCall(self, delay, func, *arg, **kw)
        call.set(self.get_advance_time())
        return call

    def get_advance_time(self) -> float:
        if not self.advance_call:
            return None
        return self.advance_call.getTime() - self.advance_call.seconds()


def run() -> None:
    """
    runs the server
    """

    # load and apply regular scripts
    script_names = scripts_option.get()
    script_dir = os.path.join(config.config_dir, 'scripts/')
    script_objects = extensions.load_scripts_regular_extension(
        script_names, script_dir)
    (protocol_class, connection_class) = extensions.apply_scripts(
        script_objects, config, FeatureProtocol, FeatureConnection)

    # load and apply the game_mode script
    game_mode_name = game_mode.get()
    game_mode_dir = os.path.join(config.config_dir, 'game_modes/')
    game_mode_object = extensions.load_script_game_mode(
        game_mode_name, game_mode_dir)
    (protocol_class, connection_class) = extensions.apply_scripts(
        game_mode_object, config, protocol_class, connection_class)

    protocol_class.connection_class = connection_class

    interface = network_interface.get().encode('utf-8')

    # instantiate the protocol class once. It will set timers and hooks to keep
    # itself running once we start the reactor
    protocol_class(interface, config.get_dict())

    log.debug('Checking for unregistered config items...')
    unused = config.check_unused()
    if unused:
        log.warn('The following config items are not used:')
        pprint(unused)

    log.info('Started server...')

    profile = logging_profile_option.get()
    if profile:
        import cProfile
        cProfile.runctx('reactor.run()', None, globals())
    else:
        reactor.run()
